/* Copyright (C) 1997-1999, Ghostgum Software Pty Ltd.  All rights reserved.
  
  This file is part of RedMon.
  
  This program is distributed with NO WARRANTY OF ANY KIND.  No author
  or distributor accepts any responsibility for the consequences of using it,
  or for whether it serves any particular purpose or works at all, unless he
  or she says so in writing.  Refer to the RedMon Free Public Licence 
  (the "Licence") for full details.
  
  Every copy of RedMon must include a copy of the Licence, normally in a 
  plain ASCII text file named LICENCE.  The Licence grants you the right 
  to copy, modify and redistribute RedMon, but only under certain conditions 
  described in the Licence.  Among other things, the Licence requires that 
  the copyright notice and this notice be preserved on all copies.
*/

/* redmon.c */

/*
 * This is a Port Monitor for Windows 95, 98 and Windows NT.
 * The monitor name is "Redirected Port" .
 * A write to any port provided by this monitor will be
 * redirected to a program using a pipe to stdin.
 *
 * An example is redirecting the output from the 
 * PostScript printer driver to Ghostscript, which then
 * writes to a non-PostScript printer.
 *
 * For efficiency reasons, don't use the C run time library. 
 *
 * The Windows NT version must use Unicode, so Windows NT
 * specific code is conditionally compiled with 
 *   #ifdef UNICODE
 *   #endif
 */


#define STRICT
#include <windows.h>
#include "redmon.h"
#ifdef BETA
#include <time.h>
#endif

typedef struct redata {
    HANDLE hPort;		/* handle to this structure */
    /* stored in registry */
    BOOL debug;
    DWORD delay;		/* delay in seconds to wait for shut down */
    DWORD show;			/* Normal, Minimized, Hidden */
    DWORD output;		/* self, stdout, temporary file */
    DWORD runuser;		/* Run as User */
    TCHAR printername[MAXSTR];	/* output printer name */
    BOOL print_error;		/* print error report */
    /* Details obtained during OpenPort  */
    TCHAR portname[MAXSHORTSTR];
    /* Details obtained during StartDocPort  */
    TCHAR command[1024];
    TCHAR pPrinterName[MAXSTR];	/* Printer name for RedMon port */
    TCHAR DocName[MAXSTR];	/* Document Name (from StartDocPort) */
    DWORD JobId;
    /* For running process */
    BOOL started;		/* true if process started */
    BOOL error;			/* true if process terminates early */
    HGLOBAL environment;	/* environment strings for process */
    HANDLE hChildStdinRd;
    HANDLE hChildStdinWr;	/* We write to this one */
    HANDLE hChildStdoutRd; 	/* We read from this one */
    HANDLE hChildStdoutWr;
    HANDLE hChildStderrRd; 	/* We read from this one */
    HANDLE hChildStderrWr;
#ifdef SAVESTD
    HANDLE hSaveStdin;
    HANDLE hSaveStdout;
    HANDLE hSaveStderr;
#endif
    HANDLE hPipeRd; 	/* We read printer output from this one */
    HANDLE hPipeWr;
    PROCESS_INFORMATION piProcInfo;
    /*  */
    TCHAR logname[MAXSTR];
    HANDLE hLogFile;
    HANDLE hmutex;	/* To control access to pipe and file handles */
    HANDLE primary_token;  	/* primary token for caller */
    /* for write thread */
    HANDLE write_event;	/* To unblock write thread */
    BOOL write;		/* TRUE if write thread should keep running */
    HANDLE write_hthread;
    DWORD write_threadid;
    LPBYTE write_buffer;	/* data to write */
    DWORD write_buffer_length;	/* number of bytes of data to write */
    BOOL write_flag;		/* TRUE if WriteFile was successful */
    DWORD write_written;	/* number of bytes written */
    /* for output to second printer queue */
    TCHAR tempname[MAXSTR];	/* temporary file name  */
    HANDLE printer;		/* handle to an printer */ 
    DWORD printer_bytes;
    BYTE pipe_buf[PIPE_BUF_SIZE]; /* buffer for use in flush_stdout */
} REDATA;


typedef struct add_param {
	LPTSTR portname;
	LPTSTR portdesc;
} ADDPARAM;

BOOL start_redirect(REDATA * prd);
void reset_redata(REDATA *prd);
BOOL check_process(REDATA *prd);
BOOL CALLBACK AddDlgProc(HWND hDlg, UINT message, 
        WPARAM wParam, LPARAM lParam);
BOOL CALLBACK ConfigDlgProc(HWND hDlg, UINT message, 
	WPARAM wParam, LPARAM lParam);
BOOL CALLBACK LogfileDlgProc(HWND hDlg, UINT message, 
	WPARAM wParam, LPARAM lParam);
int create_tempfile(LPTSTR filename, DWORD len);
int redmon_printfile(REDATA * prd, TCHAR *filename);
BOOL redmon_open_printer(REDATA *prd);
BOOL redmon_abort_printer(REDATA *prd);
BOOL redmon_close_printer(REDATA *prd);
BOOL redmon_write_printer(REDATA *prd, BYTE *ptr, DWORD len);
BOOL make_env(REDATA * prd);


/* These are our only global variables */
TCHAR rekey[MAXSTR];   /* Registry key name for our use */
HINSTANCE hdll;
TCHAR helpfile[MAXSTR];
TCHAR copyright[] = COPYRIGHT;
TCHAR version[] = VERSION;

#define PORTSNAME TEXT("Ports")
#define DESCKEY TEXT("Description")
#define COMMANDKEY TEXT("Command")
#define ARGKEY TEXT("Arguments")
#define OUTPUTKEY TEXT("Output")
#define PRINTERKEY TEXT("Printer")
#define SHOWKEY TEXT("ShowWindow")
#define RUNUSERKEY TEXT("RunUser")
#define DELAYKEY TEXT("Delay")
#define LOGUSEKEY TEXT("LogFileUse")
#define LOGNAMEKEY TEXT("LogFileName")
#define LOGDEBUGKEY TEXT("LogFileDebug")
#define PRINTERRORKEY TEXT("PrintError")
#define BACKSLASH TEXT("\\")
#define DEFAULT_DELAY 300   /* seconds */
#define MINIMUM_DELAY 15
#define PRINT_BUF_SIZE 16384

/* environment variables set for the program */
#define REDMON_PORT     TEXT("REDMON_PORT=")
#define REDMON_JOB      TEXT("REDMON_JOB=")
#define REDMON_PRINTER  TEXT("REDMON_PRINTER=")
#define REDMON_MACHINE  TEXT("REDMON_MACHINE=")
#define REDMON_USER     TEXT("REDMON_USER=")
#define REDMON_DOCNAME  TEXT("REDMON_DOCNAME=")
#define REDMON_TEMP     TEXT("TEMP=")
#define REDMON_TMP      TEXT("TMP=")

/*
 * Required functions for a Port Monitor are:
 *   AddPort
 *   AddPortEx              (NT only)
 *   ClosePort
 *   ConfigurePort
 *   DeletePort
 *   EndDocPort
 *   EnumPorts
 *   GetPrinterDataFromPort
 *   InitializeMonitor      (NT 3.51 only)
 *   InitializeMonitorEx    (95 only)
 *   InitializePrintMonitor (NT only)
 *   OpenPort
 *   ReadPort
 *   SetPortOpenTimeouts
 *   StartDocPort
 *   WritePort
 */


/* mutex used for controlling access to log file and pipe handles */
void
request_mutex(REDATA *prd)
{
    if ((prd->hmutex != NULL) && (prd->hmutex != INVALID_HANDLE_VALUE))
        WaitForSingleObject(prd->hmutex, 30000);
}

void
release_mutex(REDATA *prd)
{
    if ((prd->hmutex != NULL) && (prd->hmutex != INVALID_HANDLE_VALUE))
	ReleaseMutex(prd->hmutex);
}

#ifdef BETA
int
beta_expired(void)
{
  time_t today = time(NULL);
  struct tm *t;
  t = localtime(&today);
  if (t->tm_year+1900 < BETA_YEAR)
    return 0;
  if (t->tm_year+1900 > BETA_YEAR)
    return 1;    /* beta copy has expired */
  if (t->tm_mon+1 < BETA_MONTH)
    return 0;
  if (t->tm_mon+1 > BETA_MONTH)
    return 1;    /* beta copy has expired */
  if (t->tm_mday < BETA_DAY)
    return 0;
  return 1;    /* beta copy has expired */
}

int beta(void)
{
  if (beta_expired()) {
    TCHAR buf[MAXSTR];
    TCHAR title[MAXSTR];
    LoadString(hdll, IDS_BETAEXPIRED, buf, sizeof(buf)/sizeof(TCHAR)-1);
    LoadString(hdll, IDS_TITLE, title, sizeof(title)/sizeof(TCHAR)-1);
    MessageBox(HWND_DESKTOP, buf, title, MB_OK | MB_ICONHAND);
    return 1;
  }
  return 0;
}
#endif /* BETA */


BOOL WINAPI rEnumPorts(LPTSTR pName,DWORD Level,LPBYTE  pPorts, 
         DWORD cbBuf, LPDWORD pcbNeeded, LPDWORD pcReturned)
{
    HKEY hkey, hsubkey;
    TCHAR portname[MAXSHORTSTR], portdesc[MAXSHORTSTR];
    TCHAR monitorname[MAXSHORTSTR];
    TCHAR buf[MAXSTR];
    int needed;
    DWORD cbData, keytype;
    PORT_INFO_1 *pi1;
    PORT_INFO_2 *pi2;
    LPTSTR pstr;
    LONG rc;
    int i;

    *pcbNeeded = 0;
    *pcReturned = 0;

    if (pName != NULL) {
	return FALSE;
    }

    if ((Level < 1) || (Level > 2)) {
	SetLastError(ERROR_INVALID_LEVEL);
	return FALSE;
    }

    lstrcpy(buf, rekey);
    lstrcat(buf, BACKSLASH);
    lstrcat(buf, PORTSNAME);
    rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_READ, &hkey);
    if (rc != ERROR_SUCCESS) {
	return TRUE;	/* There are no ports */
    }

    LoadString(hdll, IDS_MONITORNAME, monitorname, 
	sizeof(monitorname)/sizeof(TCHAR)-1);

    /* First pass is to calculate the number of bytes needed */
    needed = 0;
    i = 0;
    rc = RegEnumKey(hkey, 0, portname, sizeof(portname));
    while (rc == ERROR_SUCCESS) {
	needed += (lstrlen(portname) + 1) * sizeof(TCHAR);
	if (Level == 1) {
	    needed += sizeof(PORT_INFO_1);
	}
	else if (Level == 2) {
	    needed += sizeof(PORT_INFO_2);
	    needed += (lstrlen(monitorname) + 1) * sizeof(TCHAR);
	    rc = RegOpenKeyEx(hkey, portname, 0, KEY_READ, &hsubkey);
	    if (rc == ERROR_SUCCESS) {
		cbData = sizeof(portdesc);
		keytype = REG_SZ;
		RegQueryValueEx(hsubkey, DESCKEY, 0, &keytype, 
			(LPBYTE)portdesc, &cbData);
		if (rc == ERROR_SUCCESS)
		    needed += (lstrlen(portdesc) + 1) * sizeof(TCHAR);
		else
		    needed += 1 * sizeof(TCHAR);	/* empty string */
		RegCloseKey(hsubkey);
	    }
	    else {
		needed += 1 * sizeof(TCHAR);	/* empty string */
	    }
	}
	i++;
        rc = RegEnumKey(hkey, i, portname, sizeof(portname));
    }
    *pcbNeeded = needed;
   
    if ((pPorts == NULL) || (needed > cbBuf)) {
	RegCloseKey(hkey);
	SetLastError(ERROR_INSUFFICIENT_BUFFER);
	return FALSE;
    }

    /* Second pass to copy the data to the buffer */

    /* PORT_INFO_x structures must be placed at the beginning
     * of the buffer, and strings at the end of the buffer.
     * This is important!  It appears that one buffer is 
     * allocated, then each port monitor is called in turn to
     * add its entries to the buffer, between previous entries.
     */

    i = 0;
    pi1 = (PORT_INFO_1 *)pPorts;
    pi2 = (PORT_INFO_2 *)pPorts;
    pstr = (LPTSTR)(pPorts + cbBuf);
    rc = RegEnumKey(hkey, 0, portname, sizeof(portname));
    while (rc == ERROR_SUCCESS) {
	if (Level == 1) {
	    pstr -= lstrlen(portname) + 1;
	    lstrcpy(pstr, portname);
	    pi1[i].pName = pstr;
	}
	else if (Level == 2){
	    pstr -= lstrlen(portname) + 1;
	    lstrcpy(pstr, portname);
	    pi2[i].pPortName = pstr;

	    pstr -= lstrlen(monitorname) + 1;
	    lstrcpy(pstr, monitorname);
	    pi2[i].pMonitorName = pstr;

	    rc = RegOpenKeyEx(hkey, portname, 0, KEY_READ, &hsubkey);
	    if (rc == ERROR_SUCCESS) {
		cbData = sizeof(portdesc);
		keytype = REG_SZ;
		RegQueryValueEx(hsubkey, DESCKEY, 0, &keytype, 
			(LPBYTE)portdesc, &cbData);
		if (rc != ERROR_SUCCESS)
		    portdesc[0] = '\0';
		RegCloseKey(hsubkey);
	    }
	    else {
		portdesc[0] = '\0';
	    }

	    pstr -= lstrlen(portdesc) + 1;
	    lstrcpy(pstr, portdesc);
	    pi2[i].pDescription = pstr;

	    /* Say that writing to this port is supported, */
	    /* but reading is not. */
	    /* Using fPortType = 3 is wrong. */
	    /* Options are PORT_TYPE_WRITE=1, PORT_TYPE_READ=2, */
	    /*   PORT_TYPE_REDIRECTED=4, PORT_TYPE_NET_ATTACHED=8 */
	    pi2[i].fPortType = PORT_TYPE_WRITE;
	    pi2[i].Reserved = 0;
	}
	i++;
        rc = RegEnumKey(hkey, i, portname, sizeof(portname));
    }
    *pcReturned = i;
    RegCloseKey(hkey);

    return TRUE;
}


BOOL WINAPI rAddPort(LPTSTR pName, HWND hWnd, LPTSTR pMonitorName)
{
    int i, j;
    TCHAR portname[MAXSHORTSTR], portdesc[MAXSHORTSTR];
    TCHAR buf[MAXSTR], str[MAXSTR];
    HGLOBAL hglobal;
    LPBYTE ports;
    DWORD cbBuf, needed, returned;
    PORT_INFO_2 *pi2;
    BOOL unique;
    LONG rc;
    HKEY hkey, hsubkey;
    ADDPARAM ap;

    if (pName != NULL)
	return FALSE;

    /* Get the list of existing ports. */
    /* Ask for all ports, not just ours.
     * If we ask for only RedMon ports by calling rEnumPorts, a user
     * can create a second LPT1: This isn't useful because it doesn't 
     * get serviced before the Local Port, and we can't delete it
     * with the Delete Port button.  Instead we have to use the
     * Registry Editor to clean up the mess.
     * To avoid this, stop users from entering any existing port name.
     */
    needed = 0;
    /* 1998-01-08
     * The TELES.FaxMon monitor is buggy - it returns an incorrect
     * value for "needed" if we use the following call.
     *    EnumPorts(NULL, 2, NULL, 0, &needed, &returned);
     * To dodge this bug, pass it buffer of non-zero size.
     */
   
    cbBuf = 4096;
    hglobal = GlobalAlloc(GPTR, (DWORD)cbBuf);  
    ports = GlobalLock(hglobal);
    if (ports == NULL) {
	SetLastError(ERROR_NOT_ENOUGH_MEMORY);
	return FALSE;
    }
    if (!EnumPorts(NULL, 2, ports, cbBuf, &needed, &returned)) {
	GlobalUnlock(hglobal);
	GlobalFree(hglobal);
	if (GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
	    return FALSE;	/* give up */
	}
	else {
	    /* try again, with requested size */
            cbBuf = needed;
            /* If there are no ports installed, avoid trying to 
	     * allocate 0 bytes
	     */
	    hglobal = GlobalAlloc(GPTR, (DWORD)cbBuf+4);  
	    ports = GlobalLock(hglobal);
	    if (ports == NULL) {
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);
		return FALSE;
	    }
	    if (!EnumPorts(NULL, 2, ports, cbBuf, &needed, &returned)) {
		GlobalUnlock(hglobal);
		GlobalFree(hglobal);
		return FALSE;
	    }
	}
    }

    pi2 = (PORT_INFO_2 *)ports;

    unique = TRUE;
    do {
	if (!unique) {
	    LoadString(hdll, IDS_NOTUNIQUE, str, sizeof(str)/sizeof(TCHAR)-1);
	    wsprintf(buf, str, portname);
	    LoadString(hdll, IDS_ADDPORT, str, sizeof(str)/sizeof(TCHAR)-1);
	    if (MessageBox(hWnd, buf, str, MB_OKCANCEL) == IDCANCEL) {
		GlobalUnlock(hglobal);
		GlobalFree(hglobal);
		return FALSE;
	    }
	}

	/* Suggest a unique port name */
	for (i=1; i<100; i++) {
	    wsprintf(portname, TEXT("RPT%d:"), i);
	    for (j=0; j<returned; j++) {
		if (lstrcmp(portname, pi2[j].pPortName) == 0)
		    break;
	    }
	    if (j >= returned)
		break;	/* didn't find a match so portname is unique */
	}
        LoadString(hdll, IDS_MONITORNAME, portdesc, 
		sizeof(portdesc)/sizeof(TCHAR)-1);

	ap.portname = portname;
	ap.portdesc = portdesc;
	if (!DialogBoxParam(hdll, MAKEINTRESOURCE(IDD_ADDPORT), 
		hWnd, AddDlgProc, (LPARAM)&ap)) {
	    GlobalUnlock(hglobal);
	    GlobalFree(hglobal);
	    return FALSE;
	}

	if (lstrlen(portname) && portname[lstrlen(portname)-1] != ':')
	    lstrcat(portname, TEXT(":"));	/* append ':' if not present */

	for (j=0; j<returned; j++) {
	    if (lstrcmp(portname, pi2[j].pPortName) == 0) {
		unique = FALSE;
		break;
	    }
	}
    } while (!unique);

    GlobalUnlock(hglobal);
    GlobalFree(hglobal);

    /* store new port name in registry */
    rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, rekey, 0, KEY_ALL_ACCESS, &hkey);
    if (rc == ERROR_SUCCESS) {
	lstrcpy(buf, PORTSNAME);
	lstrcat(buf, BACKSLASH);
	lstrcat(buf, portname);
	rc = RegCreateKey(hkey, buf, &hsubkey);
	if (rc == ERROR_SUCCESS) {
	    RegSetValueEx(hsubkey, DESCKEY, 0, REG_SZ,
		(CONST BYTE *)portdesc, (lstrlen(portdesc)+1)*sizeof(TCHAR));
	    RegCloseKey(hsubkey);
	}
	RegCloseKey(hkey);
    } 

   return TRUE;
}

#ifdef UNICODE
/* Windows NT only */
/* untested */
BOOL WINAPI rAddPortEx(LPWSTR pName, DWORD Level, LPBYTE lpBuffer, 
     LPWSTR lpMonitorName)
{
    int j;
    TCHAR *portname;
    TCHAR portdesc[MAXSHORTSTR];
    TCHAR buf[MAXSTR];
    HGLOBAL hglobal;
    LPBYTE ports;
    DWORD cbBuf, needed, returned;
    PORT_INFO_2 *pi2;
    LONG rc;
    HKEY hkey, hsubkey;

    if (pName != NULL)
	return FALSE;

    if ((Level != 1) && (Level != 2)) {
	SetLastError(ERROR_INVALID_LEVEL);
	return FALSE;
    }

    /* Get the list of existing ports. */
    EnumPorts(NULL, 2, NULL, 0, &needed, &returned);
    cbBuf = needed;
    hglobal = GlobalAlloc(GPTR, (DWORD)cbBuf);
    ports = GlobalLock(hglobal);
    if (ports == NULL) {
	SetLastError(ERROR_NOT_ENOUGH_MEMORY);
	return FALSE;
    }

    if (!EnumPorts(NULL, 2, ports, cbBuf, &needed, &returned)) {
	GlobalUnlock(hglobal);
	GlobalFree(hglobal);
	return FALSE;
    }
    pi2 = (PORT_INFO_2 *)ports;

    if (Level == 1) {
	PORT_INFO_1 *addpi1 = (PORT_INFO_1 *)lpBuffer;
	portname = addpi1->pName;
        LoadString(hdll, IDS_MONITORNAME, portdesc, 
		sizeof(portdesc)/sizeof(TCHAR)-1);
    }
    if (Level == 2) {
	PORT_INFO_2 *addpi2 = (PORT_INFO_2 *)lpBuffer;
	portname = addpi2->pPortName;
	lstrcpyn(portdesc, addpi2->pDescription, 
		sizeof(portdesc)/sizeof(TCHAR)-1);
    }

    /* Make sure portname is unique */
    for (j=0; j<returned; j++) {
	if (lstrcmp(portname, pi2[j].pPortName) == 0) {
	    /* fail because port already exists */
	    GlobalUnlock(hglobal);
	    GlobalFree(hglobal);
	    return FALSE;
	}
    }

    GlobalUnlock(hglobal);
    GlobalFree(hglobal);

    /* store new port name in registry */
    rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, rekey, 0, KEY_ALL_ACCESS, &hkey);
    if (rc == ERROR_SUCCESS) {
	lstrcpy(buf, PORTSNAME);
	lstrcat(buf, BACKSLASH);
	lstrcat(buf, portname);
	rc = RegCreateKey(hkey, buf, &hsubkey);
	if (rc == ERROR_SUCCESS) {
	    RegSetValueEx(hsubkey, DESCKEY, 0, REG_SZ,
		(CONST BYTE *)portdesc, (lstrlen(portdesc)+1)*sizeof(TCHAR));
	    RegCloseKey(hsubkey);
	}
	RegCloseKey(hkey);
    } 
	
   return TRUE;
}
#endif

BOOL WINAPI rDeletePort(LPTSTR pName, HWND hWnd, LPTSTR pPortName)
{
    HKEY hkey;
    LONG rc;
    TCHAR buf[MAXSTR];
    if (pName != NULL)
	return FALSE;

    lstrcpy(buf, rekey);
    lstrcat(buf, BACKSLASH);
    lstrcat(buf, PORTSNAME);
    rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_ALL_ACCESS, &hkey);
    if (rc != ERROR_SUCCESS) {
	SetLastError(REGDB_E_KEYMISSING);
	return FALSE;	/* There are no ports */
    }
    RegDeleteKey(hkey, pPortName); 
    RegCloseKey(hkey);

    return TRUE;
}


BOOL WINAPI rConfigurePort(LPTSTR pName, HWND hWnd, LPTSTR pPortName)
{
    if (pName != NULL)
	return FALSE;

    (void)DialogBoxParam(hdll, MAKEINTRESOURCE(IDD_CONFIGPORT), 
		hWnd, ConfigDlgProc, (LPARAM)pPortName);
    return TRUE;
}



BOOL WINAPI rOpenPort(LPTSTR pName, PHANDLE pHandle)
{
    HGLOBAL hglobal;
    REDATA *prd;

    TCHAR buf[MAXSTR];
#ifdef BETA
     if (beta())
	return FALSE;
#endif

    hglobal = GlobalAlloc(GPTR, (DWORD)sizeof(REDATA));
    prd = (REDATA *)GlobalLock(hglobal);
    if (prd == (REDATA *)NULL) {
	SetLastError(ERROR_NOT_ENOUGH_MEMORY);
	return FALSE;
    }
    FillMemory((PVOID)prd, sizeof(REDATA), 0);
    reset_redata(prd);
    lstrcpy(prd->portname, pName);
    prd->hPort = hglobal;

    /* Do the rest of the opening in rStartDocPort() */

    GlobalUnlock(hglobal);
    *pHandle = (HANDLE)hglobal;
    return TRUE;
}

#ifdef UNUSED
/* Don't use OpenPortEx in a Port Monitor */
BOOL WINAPI rOpenPortEx(LPTSTR  pPortName, 
        LPTSTR  pPrinterName, PHANDLE pHandle, struct _MONITOR 
        FAR *pMonitor);
#endif


/* The log file is single byte characters only */
/* Write a single character or wide character string to the log file,
 * converting it to single byte characters */
void write_string_to_log(REDATA *prd, LPTSTR buf)
{
int count;
CHAR cbuf[256];
BOOL UsedDefaultChar;
DWORD cbWritten;
    if (prd->hLogFile == INVALID_HANDLE_VALUE)
	return;

    request_mutex(prd);
#ifdef UNICODE
    while (lstrlen(buf)) {
	count = min(lstrlen(buf), sizeof(cbuf));
	WideCharToMultiByte(CP_ACP, 0, buf, count,
		cbuf, sizeof(cbuf), NULL, &UsedDefaultChar);
	buf += count;
	WriteFile(prd->hLogFile, cbuf, count, &cbWritten, NULL);
    }
#else
    WriteFile(prd->hLogFile, buf, lstrlen(buf), &cbWritten, NULL);
#endif
    FlushFileBuffers(prd->hLogFile);
    release_mutex(prd);
}

void
write_error(REDATA *prd, DWORD err)
{
LPVOID lpMessageBuffer;
    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
	FORMAT_MESSAGE_FROM_SYSTEM,
	NULL, err,
	MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), /* user default language */
	(LPTSTR) &lpMessageBuffer, 0, NULL);
    if (lpMessageBuffer) {
	write_string_to_log(prd, (LPTSTR)lpMessageBuffer);
	LocalFree(LocalHandle(lpMessageBuffer));
	write_string_to_log(prd, TEXT("\r\n"));
    }
}

void
reset_redata(REDATA *prd)
{
    prd->started = FALSE;
    prd->error = FALSE;
    prd->command[0] = '\0';
    prd->delay = DEFAULT_DELAY;
    prd->show = 0;
    prd->output = 0;
    prd->runuser = 0;
    prd->pPrinterName[0] = '\0';
    prd->DocName[0] = '\0';
    prd->JobId = 0;
    prd->hChildStdinRd = INVALID_HANDLE_VALUE;
    prd->hChildStdinWr = INVALID_HANDLE_VALUE;
    prd->hChildStdoutRd = INVALID_HANDLE_VALUE;
    prd->hChildStdoutWr = INVALID_HANDLE_VALUE;
    prd->hChildStderrRd = INVALID_HANDLE_VALUE;
    prd->hChildStderrWr = INVALID_HANDLE_VALUE;
#ifdef SAVESTD
    prd->hSaveStdin = INVALID_HANDLE_VALUE;
    prd->hSaveStdout = INVALID_HANDLE_VALUE;
    prd->hSaveStderr = INVALID_HANDLE_VALUE;
#endif
    prd->hPipeRd = INVALID_HANDLE_VALUE;
    prd->hPipeWr = INVALID_HANDLE_VALUE;
    prd->hLogFile = INVALID_HANDLE_VALUE;
    prd->piProcInfo.hProcess = INVALID_HANDLE_VALUE;
    prd->piProcInfo.hThread = INVALID_HANDLE_VALUE;
    prd->hmutex = INVALID_HANDLE_VALUE;
    prd->write = FALSE;
    prd->write_event = INVALID_HANDLE_VALUE;
    prd->write_hthread = INVALID_HANDLE_VALUE;
    prd->write_threadid = 0;
    prd->write_buffer = NULL;
    prd->write_buffer_length = 0;
    prd->tempname[0] = '\0';
    prd->printername[0] = '\0';
    prd->printer = INVALID_HANDLE_VALUE;
    prd->printer_bytes = 0;
}

/* copy stdout and stderr to log file, if open */
BOOL
flush_stdout(REDATA *prd)
{
    DWORD bytes_available, dwRead, dwWritten;
    BOOL result;
    BOOL got_something = FALSE;

    request_mutex(prd);

    /* copy anything on stdout to printer or log file */
    bytes_available = 0;
    result = PeekNamedPipe(prd->hChildStdoutRd, NULL, 0, NULL, 
	    &bytes_available, NULL);
    while (result && bytes_available) {
	if (!ReadFile(prd->hChildStdoutRd, prd->pipe_buf, sizeof(prd->pipe_buf), 
	    &dwRead, NULL) || dwRead == 0) 
	    break;
	got_something = TRUE;
	if (prd->output == OUTPUT_STDOUT) {
	    if (prd->printer != INVALID_HANDLE_VALUE) {
		if (!redmon_write_printer(prd, prd->pipe_buf, dwRead)) {
		    redmon_abort_printer(prd);
		}
	    }
	}
	else if (prd->hLogFile != INVALID_HANDLE_VALUE) {
	    WriteFile(prd->hLogFile, prd->pipe_buf, dwRead, &dwWritten, NULL);
	    FlushFileBuffers(prd->hLogFile);
	}
	result = PeekNamedPipe(prd->hChildStdoutRd, NULL, 0, NULL, 
	    &bytes_available, NULL);
    }

    /* copy anything on stderr to log file */
    bytes_available = 0;
    result = PeekNamedPipe(prd->hChildStderrRd, NULL, 0, NULL, 
	    &bytes_available, NULL);
    while (result && bytes_available) {
	if (!ReadFile(prd->hChildStderrRd, prd->pipe_buf, sizeof(prd->pipe_buf), &dwRead, NULL) ||
	    dwRead == 0) break;
	got_something = TRUE;
	if (prd->hLogFile != INVALID_HANDLE_VALUE) {
	    WriteFile(prd->hLogFile, prd->pipe_buf, dwRead, &dwWritten, NULL);
	    FlushFileBuffers(prd->hLogFile);
	}
	result = PeekNamedPipe(prd->hChildStderrRd, NULL, 0, NULL, 
	    &bytes_available, NULL);
    }

    /* copy anything on printer pipe to the printer */
    if (prd->output == OUTPUT_HANDLE) {
	bytes_available = 0;
	result = PeekNamedPipe(prd->hPipeRd, NULL, 0, NULL, 
		&bytes_available, NULL);
	while (result && bytes_available) {
	    if (!ReadFile(prd->hPipeRd, prd->pipe_buf, sizeof(prd->pipe_buf), 
		&dwRead, NULL) || dwRead == 0) 
		break;
	    got_something = TRUE;
	    if (prd->printer != INVALID_HANDLE_VALUE) {
		if (!redmon_write_printer(prd, prd->pipe_buf, dwRead)) {
		    redmon_abort_printer(prd);
		}
	    }
	    result = PeekNamedPipe(prd->hPipeRd, NULL, 0, NULL, 
		&bytes_available, NULL);
	}
    }

    release_mutex(prd);
    return got_something;
}


/* Check if process is running.  */
/* Return TRUE if process is running, FALSE otherwise */
/* Shut down stdin pipe if we find process has terminated */
BOOL check_process(REDATA *prd)
{
    DWORD exit_status;
    if (prd->error)
	return FALSE;	/* process is not running */

    if (prd->piProcInfo.hProcess == INVALID_HANDLE_VALUE)
	prd->error = TRUE;
    if (!prd->error 
	&& GetExitCodeProcess(prd->piProcInfo.hProcess, &exit_status)
	&& (exit_status != STILL_ACTIVE))
	prd->error = TRUE;

    if (prd->error) {
	DWORD bytes_available, dwRead;
	BOOL result;
	BYTE buf[256];
	if (prd->debug) {
	  write_string_to_log(prd, 
	    TEXT("REDMON check_process: process isn't running.\r\n"));
	  write_string_to_log(prd, 
	    TEXT("REDMON check_process: flushing child stdin to unblock WriteThread.\r\n"));
	}
	/* flush stdin pipe to unblock WriteThread */
	bytes_available = 0;
	result = PeekNamedPipe(prd->hChildStdinRd, NULL, 0, NULL, 
		    &bytes_available, NULL);
	while (result && bytes_available) {
	    ReadFile(prd->hChildStdinRd, buf, sizeof(buf), &dwRead, NULL);
	    result = PeekNamedPipe(prd->hChildStdinRd, NULL, 0, NULL, 
		&bytes_available, NULL);
	}

    }
    return !prd->error;
}

/* Thread to write to stdout pipe */
DWORD WINAPI WriteThread(LPVOID lpThreadParameter)
{
    HANDLE hPort = (HANDLE)lpThreadParameter;
    REDATA *prd = GlobalLock((HGLOBAL)hPort);
    

    if (prd == (REDATA *)NULL)
	return 1;

    if (prd->debug)
	write_string_to_log(prd, TEXT("\r\nREDMON WriteThread: started\r\n"));

    while (prd->write && !prd->error) {
	WaitForSingleObject(prd->write_event, INFINITE);
	ResetEvent(prd->write_event);
	if (prd->write_buffer_length && prd->write_buffer) {
	    if (! (prd->write_flag = WriteFile(prd->hChildStdinWr, 
		prd->write_buffer, prd->write_buffer_length, 
		&prd->write_written, NULL)) )
		prd->write = FALSE;	/* get out of here */
	}
	prd->write_buffer = NULL;
	prd->write_buffer_length = 0;
    }

    CloseHandle(prd->write_event);
    prd->write_event = INVALID_HANDLE_VALUE;

    if (prd->debug)
	write_string_to_log(prd, TEXT("\r\nREDMON WriteThread: ending\r\n"));

    GlobalUnlock(hPort);
    return 0;
}


BOOL WINAPI rStartDocPort(HANDLE  hPort, LPTSTR  pPrinterName, 
        DWORD   JobId, DWORD   Level, LPBYTE  pDocInfo)
{
    TCHAR buf[MAXSTR];
    TCHAR args[MAXSTR];
    int i;
    LPTSTR s, d;
    HKEY hkey;
    REDATA *prd = GlobalLock((HGLOBAL)hPort);
    BOOL flag;
    LONG rc;
    DWORD cbData, keytype;
    DWORD uselog;

    if (prd == (REDATA *)NULL) {
	SetLastError(ERROR_INVALID_HANDLE);
	return FALSE;
    }

    reset_redata(prd);
    lstrcpy(prd->pPrinterName, pPrinterName);
    prd->JobId = JobId;
    /* remember document name, to be used for output job */
    if ((Level == 1) && pDocInfo) {
	DOC_INFO_1 *dci1 = (DOC_INFO_1 *)pDocInfo;
	lstrcpyn(prd->DocName, dci1->pDocName, 
	    sizeof(prd->DocName)/sizeof(TCHAR));
    }
    else if ((Level == 2) && pDocInfo) {
	DOC_INFO_2 *dci2 = (DOC_INFO_2 *)pDocInfo;
	lstrcpyn(prd->DocName, dci2->pDocName, 
	    sizeof(prd->DocName)/sizeof(TCHAR));
    }
    else
	lstrcpy(prd->DocName, TEXT("RedMon"));

    /* Launch application */

    lstrcpy(buf, rekey);
    lstrcat(buf, BACKSLASH);
    lstrcat(buf, PORTSNAME);
    lstrcat(buf, BACKSLASH);
    lstrcat(buf, prd->portname);
    rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_READ, &hkey);
    if (rc != ERROR_SUCCESS) {
	GlobalUnlock((HGLOBAL)hPort);
	SetLastError(REGDB_E_KEYMISSING);
	return FALSE;	/* There are no ports */
    }

    /* query if log is to be used */
    cbData = sizeof(DWORD);
    keytype = REG_DWORD;
    uselog = 0;
    RegQueryValueEx(hkey, LOGUSEKEY, 0, &keytype, 
	(LPBYTE)&uselog, &cbData);
    if (uselog) {
	/* Open optional log file */
	prd->hLogFile = INVALID_HANDLE_VALUE;
	prd->logname[0] = '\0';
	cbData = sizeof(prd->logname);
	keytype = REG_SZ;
	RegQueryValueEx(hkey, LOGNAMEKEY, 0, &keytype, 
		(LPBYTE)prd->logname, &cbData);
	if (lstrlen(prd->logname)) {
	    prd->hLogFile = CreateFile(prd->logname, GENERIC_WRITE, 
		FILE_SHARE_READ, NULL, CREATE_ALWAYS, 
		FILE_ATTRIBUTE_NORMAL, NULL);
	}

	/* find out if debug messages are required */
	cbData = sizeof(DWORD);
	keytype = REG_DWORD;
	RegQueryValueEx(hkey, LOGDEBUGKEY, 0, &keytype, 
	    (LPBYTE)&prd->debug, &cbData);
	if (prd->debug) {
	    LoadString(hdll, IDS_TITLE, buf, sizeof(buf)/sizeof(TCHAR)-1);
	    write_string_to_log(prd, buf);
	    write_string_to_log(prd, TEXT("\r\n"));
	    write_string_to_log(prd, copyright);
	    write_string_to_log(prd, version);
	}
    }

    flag = TRUE;	/* all is well */

    /* Find out which output method is used */
    cbData = sizeof(DWORD);
    keytype = REG_DWORD;
    RegQueryValueEx(hkey, OUTPUTKEY, 0, &keytype, 
	(LPBYTE)&prd->output, &cbData);

    if (prd->output == OUTPUT_FILE) {
	if (!create_tempfile(prd->tempname, 
		sizeof(prd->tempname)/sizeof(TCHAR))) {
	    write_string_to_log(prd, 
		TEXT("\r\nREDMON StartDocPort: temp file creation failed\r\n"));
	    flag = FALSE;
	}
    }
    if (prd->output != OUTPUT_SELF) {
	/* get the printer name */
	cbData = sizeof(prd->printername) - (1 * sizeof(TCHAR));
	keytype = REG_SZ;
	RegQueryValueEx(hkey, PRINTERKEY, 0, &keytype, 
	    (LPBYTE)(prd->printername), &cbData);
    }

    /* Create anonymous inheritable pipe for printer output */
    if (prd->output == OUTPUT_HANDLE) {
	HANDLE hPipeTemp;
	SECURITY_ATTRIBUTES saAttr;
	/* Set the bInheritHandle flag so pipe handles are inherited. */
	saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
	saAttr.bInheritHandle = TRUE;
	saAttr.lpSecurityDescriptor = NULL;
        if (!CreatePipe(&hPipeTemp, &prd->hPipeWr, &saAttr, 0)) {
	    write_string_to_log(prd, 
		TEXT("\r\nREDMON StartDocPort: open printer pipe failed\r\n"));
	    flag = FALSE;
	}
	/* make the read handle non-inherited */
	if (!DuplicateHandle(GetCurrentProcess(), hPipeTemp,
		GetCurrentProcess(), &prd->hPipeRd, 0,
		FALSE,       /* not inherited */
		DUPLICATE_SAME_ACCESS)) {
	    write_string_to_log(prd, 
		TEXT("\r\nREDMON StartDocPort: duplicate printer pipe handle failed\r\n"));
	    flag = FALSE;
	}
	CloseHandle(hPipeTemp);
    }
  
    if (!flag) {
	RegCloseKey(hkey);
	if (prd->hLogFile != INVALID_HANDLE_VALUE);
	    CloseHandle(prd->hLogFile);
	prd->hLogFile = INVALID_HANDLE_VALUE;
	if (prd->hPipeRd != INVALID_HANDLE_VALUE);
	    CloseHandle(prd->hPipeRd);
	prd->hPipeRd = INVALID_HANDLE_VALUE;
	if (prd->hPipeWr != INVALID_HANDLE_VALUE);
	    CloseHandle(prd->hPipeWr);
	prd->hPipeWr = INVALID_HANDLE_VALUE;
	GlobalUnlock((HGLOBAL)hPort);
	return FALSE;
    }

    /* Build command line */
    lstrcpy(prd->command, TEXT("\042"));
    cbData = sizeof(prd->command) - (1 * sizeof(TCHAR));
    keytype = REG_SZ;
    RegQueryValueEx(hkey, COMMANDKEY, 0, &keytype, 
	(LPBYTE)(prd->command+1), &cbData);
    lstrcat(prd->command, TEXT("\042 "));

    cbData = sizeof(args) - (1 * sizeof(TCHAR));
    RegQueryValueEx(hkey, ARGKEY, 0, &keytype, (LPBYTE)(args), &cbData);

    /* copy arguments, substituting %1 for temp, %h for printer pipe handle, */
    /* %d for document name */
    i = lstrlen(prd->command);
    for (s = args; *s && (i < sizeof(prd->command)/sizeof(TCHAR)-1); s++) {
	if ( (*s == '%') && (*(s+1)=='1') &&
	  (i+lstrlen(prd->tempname) < sizeof(prd->command)/sizeof(TCHAR)-1) )
	{
	    /* copy temp filename */
	    prd->command[i] = '\0';
	    lstrcat(prd->command, prd->tempname);
	    i = lstrlen(prd->command);
	    s++;
        }
	else if ( (*s == '%') && (*(s+1)=='h') &&
	  (i+16 < sizeof(prd->command)/sizeof(TCHAR)-1) )
	{
	    /* copy printer pipe handle as hexadecimal */
	    prd->command[i] = '\0';
	    wsprintf(&(prd->command[i]), TEXT("%08x"), (int)(prd->hPipeWr));
	    i = lstrlen(prd->command);
	    s++;
        }
	else if ( (*s == '%') && (*(s+1)=='d') &&
	  (i+lstrlen(prd->DocName) < sizeof(prd->command)/sizeof(TCHAR)-1) )
	{
	    /* copy document name */
	    prd->command[i] = '\0';
	    lstrcat(prd->command, prd->DocName);
	    i = lstrlen(prd->command);
	    s++;
        }
	else if ( (*s == '%') && (*(s+1)=='%') ) {
	    s++;
	    prd->command[i++] = *s;
	}
	else
	    prd->command[i++] = *s;
    }
    prd->command[i] = '\0';

    /* Find out if window is Normal, Minimized or Hidden */
    cbData = sizeof(DWORD);
    keytype = REG_DWORD;
    RegQueryValueEx(hkey, SHOWKEY, 0, &keytype, 
	(LPBYTE)&prd->show, &cbData);

    /* set shutdown delay */
    cbData = sizeof(DWORD);
    keytype = REG_DWORD;
    RegQueryValueEx(hkey, DELAYKEY, 0, &keytype, 
	(LPBYTE)&prd->delay, &cbData);
    if (prd->delay < MINIMUM_DELAY)
	prd->delay = MINIMUM_DELAY;

#if defined(UNICODE) && !defined(NT35)
    /* Find out if NT4 should impersonate user */
    /* CreateProcessAsUser is only supported by NT3.51 and later. */
    cbData = sizeof(DWORD);
    keytype = REG_DWORD;
    RegQueryValueEx(hkey, RUNUSERKEY, 0, &keytype, 
	(LPBYTE)&prd->runuser, &cbData);
#endif

    /* find out if error report is to be printed */
    cbData = sizeof(DWORD);
    keytype = REG_DWORD;
    prd->print_error = 0;
    RegQueryValueEx(hkey, PRINTERRORKEY, 0, &keytype, 
	(LPBYTE)&prd->print_error, &cbData);

    RegCloseKey(hkey);

    if ( (prd->output == OUTPUT_STDOUT) ||
	 (prd->output == OUTPUT_HANDLE) ) {
	/* open a printer */
	if (!redmon_open_printer(prd)) {
	    write_string_to_log(prd, 
		TEXT("\r\nREDMON StartDocPort: open printer failed\r\n"));
	    if (prd->hLogFile != INVALID_HANDLE_VALUE)
	        CloseHandle(prd->hLogFile);
	    prd->hLogFile = INVALID_HANDLE_VALUE;
	    GlobalUnlock((HGLOBAL)hPort);
	    return FALSE;
	}
    }

    make_env(prd);

    prd->hmutex = CreateMutex(NULL, FALSE, NULL);
    flag = start_redirect(prd);
    if (flag) {
        WaitForInputIdle(prd->piProcInfo.hProcess, 5000);

	/* Create thread to write to stdin pipe
	 * We need this to avoid a deadlock when stdin and stdout
	 * pipes are both blocked.
	 */
	prd->write_event = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (prd->write_event == NULL)
	    write_string_to_log(prd, 
		TEXT("couldn't create synchronization event\r\n"));
	prd->write = TRUE;
	prd->write_hthread = CreateThread(NULL, 0, &WriteThread, 
		hPort, 0, &prd->write_threadid);
    }
    else {
	DWORD cbWritten;
	DWORD err = GetLastError();
	/* ENGLISH */
	if (prd->environment) {
	    GlobalUnlock(prd->environment);
	    GlobalFree(prd->environment);
	    prd->environment = NULL;
	}
	wsprintf(buf, 
	   TEXT("StartDocPort: failed to start process\r\n  Port = %s\r\n  Command = %s\r\n  Error = %ld\r\n"),
	   prd->portname, prd->command, err);
	switch(err) {
	    case ERROR_FILE_NOT_FOUND:
		lstrcat(buf, TEXT("  File not found\r\n"));
		break;
	    case ERROR_PATH_NOT_FOUND:
		lstrcat(buf, TEXT("  Path not found\r\n"));
		break;
	    case ERROR_BAD_PATHNAME:
		lstrcat(buf, TEXT("  Bad path name\r\n"));
		break;
	}
	write_string_to_log(prd, buf);
	write_error(prd, err);
    }

    if (prd->debug) {
	wsprintf(buf, 
	  TEXT("REDMON StartDocPort: returning %d\r\n\
  %s\r\n\
  Printer=%s\r\n\
  JobId=%d\r\n"), 
	  flag, prd->command, prd->pPrinterName, prd->JobId);
	write_string_to_log(prd, buf);
	if ((Level == 1) && pDocInfo) {
	    DOC_INFO_1 *dci1 = (DOC_INFO_1 *)pDocInfo;
	    wsprintf(buf, TEXT("\
  Level=1\r\n\
    DocumentName=\042%s\042\r\n\
    OutputFile=\042%s\042\r\n\
    Datatype=\042%s\042\r\n"),
		dci1->pDocName ? dci1->pDocName : TEXT("(null)"), 
		dci1->pOutputFile ? dci1->pOutputFile : TEXT("(null)"),
		dci1->pDatatype ? dci1->pDatatype : TEXT("(null)"));
	    write_string_to_log(prd, buf);
	}
	else if ((Level == 2) && pDocInfo) {
	    DOC_INFO_2 *dci2 = (DOC_INFO_2 *)pDocInfo;
	    wsprintf(buf, TEXT("\
  Level=2\r\n\
    DocumentName=\042%s\042\r\n\
    OutputFile=\042%s\042\r\n\
    Datatype=\042%s\042\r\n\
    Mode=%d\r\n\
    JobId=%d\r\n"),
		dci2->pDocName ? dci2->pDocName : TEXT("(null)"), 
		dci2->pOutputFile ? dci2->pOutputFile : TEXT("(null)"),
		dci2->pDatatype ? dci2->pDatatype : TEXT("(null)"),
		dci2->dwMode, dci2->JobId);
	    write_string_to_log(prd, buf);
	}
	else {
	    wsprintf(buf, TEXT("  Level=%d pDocInfo=%d\r\n"),
		Level, pDocInfo);
	    write_string_to_log(prd, buf);
	}
	wsprintf(buf, TEXT("  output=%d show=%d delay=%d runuser=%d\r\n"),
	    prd->output, prd->show, prd->delay, prd->runuser);
	write_string_to_log(prd, buf);
    }

    if (!flag) {
	/* close all file and object handles */
	if (prd->hLogFile != INVALID_HANDLE_VALUE);
	    CloseHandle(prd->hLogFile);
	prd->hLogFile = INVALID_HANDLE_VALUE;

	if (prd->hChildStderrRd)
	    CloseHandle(prd->hChildStderrRd);
	if (prd->hChildStderrWr)
	    CloseHandle(prd->hChildStderrWr);
	if (prd->hChildStdoutRd)
	    CloseHandle(prd->hChildStdoutRd);
	if (prd->hChildStdoutWr)
	    CloseHandle(prd->hChildStdoutWr);
	if (prd->hChildStdinRd)
	    CloseHandle(prd->hChildStdinRd);
	if (prd->hChildStdinWr)
	    CloseHandle(prd->hChildStdinWr);
	if (prd->hPipeRd)
	    CloseHandle(prd->hPipeRd);
	if (prd->hPipeWr)
	    CloseHandle(prd->hPipeWr);
	prd->hChildStderrRd   = INVALID_HANDLE_VALUE;
	prd->hChildStderrWr   = INVALID_HANDLE_VALUE;
	prd->hChildStdoutRd   = INVALID_HANDLE_VALUE;
	prd->hChildStdoutWr   = INVALID_HANDLE_VALUE;
	prd->hChildStdinRd    = INVALID_HANDLE_VALUE;
	prd->hChildStdinWr    = INVALID_HANDLE_VALUE;
	prd->hPipeRd   = INVALID_HANDLE_VALUE;
	prd->hPipeWr   = INVALID_HANDLE_VALUE;
#ifdef SAVESTD
	SetStdHandle(STD_INPUT_HANDLE, prd->hSaveStdin);
	SetStdHandle(STD_OUTPUT_HANDLE, prd->hSaveStdout);
	SetStdHandle(STD_ERROR_HANDLE, prd->hSaveStdout);
#endif

	if (prd->hmutex != INVALID_HANDLE_VALUE)
	    CloseHandle(prd->hmutex);
	prd->hmutex    = INVALID_HANDLE_VALUE;
    }

    prd->started = flag;

    GlobalUnlock((HGLOBAL)hPort);

    return flag;
}

void
redmon_cancel_job(REDATA *prd)
{
    TCHAR buf[MAXSTR];
    HANDLE hPrinter;
    if (OpenPrinter(prd->pPrinterName, &hPrinter, NULL)) {
	DWORD dwNeeded = 0;
	HGLOBAL hglobal = GlobalAlloc(GPTR, (DWORD)4096);
	JOB_INFO_1 *pjob = (JOB_INFO_1 *)GlobalLock(hglobal);
	if ((pjob != (JOB_INFO_1 *)NULL) && 
	     GetJob(hPrinter, prd->JobId, 1, (LPBYTE)pjob, 
		4096, &dwNeeded)) {
	    pjob->Status = JOB_STATUS_ERROR;
	    SetJob(hPrinter, prd->JobId, 1, (LPBYTE)pjob, 
		JOB_CONTROL_CANCEL);
	    if (prd->debug && (prd->hLogFile != INVALID_HANDLE_VALUE)) {
		wsprintf(buf, 
	  TEXT("\r\nREDMON WritePort: Cancelling print job method 1\r\n"));
		write_string_to_log(prd, buf);
	    }
	}
	else {
	    SetJob(hPrinter, prd->JobId, 0, NULL, JOB_CONTROL_CANCEL);
	    if (prd->debug && (prd->hLogFile != INVALID_HANDLE_VALUE)) {
		wsprintf(buf, 
	  TEXT("\r\nREDMON WritePort: Cancelling print job method 2\r\n"));
		write_string_to_log(prd, buf);
	    }
	}
	if (pjob != (JOB_INFO_1 *)NULL) {
	    GlobalUnlock(hglobal);
	    GlobalFree(hglobal);
	}
	ClosePrinter(hPrinter);
    }
}

/* WritePort is normally called between StartDocPort and EndDocPort,
 * but can be called outside this pair for bidirectional printers.
 */
BOOL WINAPI rWritePort(HANDLE  hPort, LPBYTE  pBuffer, 
        DWORD   cbBuf, LPDWORD pcbWritten)
{
    TCHAR buf[MAXSTR];
    DWORD dwRead, dwWritten;
    DWORD exit_status;
    REDATA *prd = GlobalLock((HGLOBAL)hPort);
    unsigned int sleep_count;

    if (prd == (REDATA *)NULL) {
	SetLastError(ERROR_INVALID_HANDLE);
	return FALSE;
    }

    *pcbWritten = 0;

    if (!prd->started) {
	if (prd->debug) {
	    wsprintf(buf, 
	      TEXT("REDMON WritePort: called outside Start/EndDocPort.  Returning FALSE\r\n"));
	    MessageBox(NULL, buf, MONITORNAME, MB_OK);
	}
	GlobalUnlock((HGLOBAL)hPort);
	return FALSE;
    }

    /* copy from output pipes to printer or log file */
    flush_stdout(prd);

    /* Make sure process is still running */
    check_process(prd);

    if (prd->error) {
	/* The process is no longer running, probably due to an error. */
	/* If we return an error from WritePort, the spooler crashes.
	 * To avoid this mess, don't return an error from WritePort.
	 * Instead carry on as if the error didn't occur.
	 * The only evidence of an error will be the log file.
	 */
	*pcbWritten = cbBuf;	/* say we wrote it all */
	if (prd->debug && (prd->hLogFile != INVALID_HANDLE_VALUE)) {
	    DWORD cbWritten;
	    wsprintf(buf, 
	      TEXT("\r\nREDMON WritePort: Process not running. \
Returning TRUE.\r\n    Ignoring %d bytes\r\n"), cbBuf);
	    write_string_to_log(prd, buf);
	}

	/* Cancel the print job */
	redmon_cancel_job(prd);

	GlobalUnlock((HGLOBAL)hPort);
	return TRUE;	/* say we wrote it all */
    }

    if (prd->debug && (prd->hLogFile != INVALID_HANDLE_VALUE)) {
	DWORD cbWritten;
	wsprintf(buf, 
	 TEXT("\r\nREDMON WritePort: about to write %d bytes to port.\r\n"),
	 cbBuf);
	write_string_to_log(prd, buf);
    }


    /* write to stdin pipe */
    prd->write_buffer_length = cbBuf;
    prd->write_buffer = pBuffer;
    prd->write_flag = TRUE;
    prd->write_written = 0;
    SetEvent(prd->write_event);
    if (prd->write && prd->write_buffer_length) {
	flush_stdout(prd);
	check_process(prd);
	Sleep(0);
    }
    sleep_count = 0;
    while (!prd->error && prd->write && prd->write_buffer_length) {
	/* wait for it to be written, while flushing stdout */
	if (flush_stdout(prd)) {
	    /* We succeeded in reading something from one of the
	     * pipes and the pipes are now empty.  Give up the
	     * remainder of our time slice to allow the
	     * other process to write something to the pipes
	     * or to read from stdin.
	     */
	    sleep_count = 0;
	    Sleep(0);
	}
	else if (prd->write_buffer_length) {
	    /* The pipes were empty, and the other process
	     * hasn't finished reading stdin.
	     * Pause a little until something is available or
	     * until stdin has been read.
	     * If the process is very slow and doesn't read stdin 
	     * within a reasonable time, sleep for 100ms to avoid 
	     * wasting CPU.
	     */
	    if (sleep_count < 10)
		Sleep(sleep_count * 5);
	    else
	        Sleep(100);
	    sleep_count++;
	}
	/* Make sure process is still running */
	check_process(prd);
    }
    *pcbWritten = prd->write_written;

    if (prd->error)
        *pcbWritten = cbBuf;

    if (prd->debug && (prd->hLogFile != INVALID_HANDLE_VALUE)) {
	DWORD cbWritten;
	request_mutex(prd);
        WriteFile(prd->hLogFile, pBuffer, cbBuf, &cbWritten, NULL);
	FlushFileBuffers(prd->hLogFile);
	release_mutex(prd);
	wsprintf(buf, 
	  TEXT("\r\nREDMON WritePort: %s  count=%d written=%d\r\n"), 
	      (prd->write_flag ? TEXT("OK") : TEXT("Failed")),
	      cbBuf, *pcbWritten);
	write_string_to_log(prd, buf);
    }

    flush_stdout(prd);

    if (prd->error) {
	if (prd->debug && (prd->hLogFile != INVALID_HANDLE_VALUE)) {
	    DWORD cbWritten;
	    wsprintf(buf, 
	      TEXT("\r\nREDMON WritePort: Process not running. \
Returning TRUE.\r\n    Ignoring %d bytes\r\n"), cbBuf);
	    write_string_to_log(prd, buf);
	}
	/* Cancel the print job */
	redmon_cancel_job(prd);
    }

    GlobalUnlock((HGLOBAL)hPort);

    return TRUE;	/* returning FALSE crashes Win95 spooler */
}

/* ReadPort can be called within a Start/EndDocPort pair,
 * and also outside this pair.
 */
BOOL WINAPI rReadPort(HANDLE hPort, LPBYTE pBuffer, 
        DWORD  cbBuffer, LPDWORD pcbRead)
{
    REDATA *prd = GlobalLock((HGLOBAL)hPort);
    DWORD exit_status;

    /* we don't support reading */
    *pcbRead = 0;

    if (prd == (REDATA *)NULL) {
	SetLastError(ERROR_INVALID_HANDLE);
	return FALSE;
    }

    Sleep(1000);	/* Pause a little */

    if (prd->debug) {
	TCHAR buf[MAXSTR];
	wsprintf(buf, TEXT("REDMON ReadPort: returning TRUE. Process %s\r\n"),
	   prd->error ? TEXT("has an ERROR.") : TEXT("is OK."));
	write_string_to_log(prd, buf);
	wsprintf(buf, TEXT("REDMON ReadPort: You must disable bi-directional printer support for this printer\r\n"));
	write_string_to_log(prd, buf);
    }

    GlobalUnlock((HGLOBAL)hPort);

    return TRUE;
}

BOOL WINAPI rEndDocPort(HANDLE hPort)
{
    TCHAR buf[MAXSTR];
    BOOL flag;
    DWORD bytes_available;
    DWORD dwRead, dwWritten;
    REDATA *prd = GlobalLock((HGLOBAL)hPort);
    DWORD exit_status;
    HANDLE hPrinter;
    int i;

    if (prd == (REDATA *)NULL) {
	SetLastError(ERROR_INVALID_HANDLE);
	return FALSE;
    }

    if (prd->debug)
	write_string_to_log(prd, 
		TEXT("REDMON EndDocPort: starting\r\n"));

    /* tell write thread to shut down */
    prd->write_buffer_length = 0;
    prd->write_buffer = NULL;
    prd->write = FALSE;
    SetEvent(prd->write_event);
    Sleep(0);	/* let write thread terminate */

    for (i=0; i<20; i++) {
	if (GetExitCodeThread(prd->write_hthread, &exit_status)) {
	    if (exit_status != STILL_ACTIVE)
		break;
	}
	else
	    break;
	Sleep(50);
    }
    CloseHandle(prd->write_hthread);

    /* Close stdin to signal EOF */
    if (prd->hChildStdinWr != INVALID_HANDLE_VALUE)
	CloseHandle(prd->hChildStdinWr);

    flush_stdout(prd);

    /* wait here for up to 'delay' seconds until process ends */
    /* so that process has time to write stdout/err */
    exit_status = 0;
    for (i=0; i<prd->delay; i++) {
	if (prd->piProcInfo.hProcess == INVALID_HANDLE_VALUE)
	    break;
	if (GetExitCodeProcess(prd->piProcInfo.hProcess, &exit_status)) {
	    if (exit_status != STILL_ACTIVE)
	        break;
	}
	else {
	    /* process doesn't exist */
	    break;
	}

        flush_stdout(prd);

        Sleep(1000);
    }

    flush_stdout(prd);

    if (prd->debug) {
	wsprintf(buf, 
	    TEXT("REDMON EndDocPort: process %s after %d second%s\r\n"), 
	    (exit_status == STILL_ACTIVE) ? 
		TEXT("still running") : TEXT("finished"), 
	    i, (i != 1) ? TEXT("s") : TEXT(""));
	write_string_to_log(prd, buf);
    }

    /* if process still running, we might want to kill it
     * with TerminateProcess(), but since this can have bad
     * side effects, don't do it yet.
     */

    /* When the process terminates, it will close its copy of
     * of the stdout & stderr pipe write handles. 
     * The pipes remain open because we still have a handle to
     * to the write end. 
     * Close our copy of the write end of the stdout & stderr 
     * pipes to flush them.
     */
    CloseHandle(prd->hChildStdoutWr);
    CloseHandle(prd->hChildStderrWr);
    if (prd->hPipeWr != INVALID_HANDLE_VALUE);
        CloseHandle(prd->hPipeWr);

    /* copy anything on stdout/err to log file */
    flush_stdout(prd);

    /* Close the read end of the stdio pipes */
    CloseHandle(prd->hChildStderrRd);
    CloseHandle(prd->hChildStdoutRd);
    CloseHandle(prd->hChildStdinRd);
    if (prd->hPipeWr != INVALID_HANDLE_VALUE);
        CloseHandle(prd->hPipeRd);

    /* NT documentation says *we* should cancel the print job. */
    /* 95 documentation says nothing about this. */
    if (OpenPrinter(prd->pPrinterName, &hPrinter, NULL)) {
	SetJob(hPrinter, prd->JobId, 0, NULL, JOB_CONTROL_CANCEL);
	ClosePrinter(hPrinter);
    }

#ifdef UNUSED
    /* elsewhere, the NT documentation says we should use */
    /* JOB_CONTROL_SENT_TO_PRINTER */
    if (OpenPrinter(prd->pPrinterName, &hPrinter, NULL)) {
	SetJob(hPrinter, prd->JobId, 0, NULL, JOB_CONTROL_SENT_TO_PRINTER);
	ClosePrinter(hPrinter);
    }
#endif

    /* If we output to a temporary file, print it now */
    if (prd->output == OUTPUT_FILE) {
	if (prd->debug) {
	    wsprintf(buf, 
		TEXT("\r\nREDMON EndDocPort: Copying %s to %s"), 
		   prd->tempname, prd->printername);
	    write_string_to_log(prd, buf);
	}
	redmon_printfile(prd, prd->tempname); 
	if (prd->debug)
	    write_string_to_log(prd, 
		TEXT("\r\nREDMON EndDocPort: Deleting temporary file"));
	DeleteFile(prd->tempname);
    }
    else if ( (prd->output == OUTPUT_STDOUT) || 
	(prd->output == OUTPUT_HANDLE)  ){
	redmon_close_printer(prd);
	if (prd->debug)
	  write_string_to_log(prd, 
	    TEXT("\r\nREDMON EndDocPort: Closing printer for stdout or pipe"));
    }

    if (prd->debug) {
	TCHAR buf[256];
	wsprintf(buf, 
	    TEXT("\r\nREDMON EndDocPort: %d bytes written to printer\r\n"), 
	    prd->printer_bytes);
	write_string_to_log(prd, buf);
    }

    if (prd->environment) {
	/* Are we responsible for this, or does the process delete
	 * its own environment? */
	GlobalUnlock(prd->environment);
	GlobalFree(prd->environment);
	prd->environment = NULL;
    }

    if (prd->runuser)
	CloseHandle(prd->primary_token);

    if (prd->debug)
	write_string_to_log(prd, 
		TEXT("REDMON EndDocPort: ending\r\n"));

    if (prd->hLogFile != INVALID_HANDLE_VALUE)
	CloseHandle(prd->hLogFile);

    if (prd->error && prd->print_error && (prd->output != OUTPUT_SELF))
	redmon_print_error(prd);

    if ((prd->hmutex != NULL) && (prd->hmutex != INVALID_HANDLE_VALUE))
	CloseHandle(prd->hmutex);

    if (prd->piProcInfo.hProcess != INVALID_HANDLE_VALUE)
	CloseHandle(prd->piProcInfo.hProcess);

    if (prd->piProcInfo.hThread != INVALID_HANDLE_VALUE)
	CloseHandle(prd->piProcInfo.hThread);

    reset_redata(prd);

    GlobalUnlock((HGLOBAL)hPort);

    return TRUE;
}


BOOL WINAPI rClosePort(HANDLE  hPort)
{
    TCHAR buf[MAXSTR];

    /* assume files were all closed in rEndDocPort() */

    if (hPort)
	GlobalFree((HGLOBAL)hPort);

    return TRUE;
}


BOOL WINAPI rSetPortTimeOuts(HANDLE  hPort, LPCOMMTIMEOUTS lpCTO, DWORD reserved)
{
    TCHAR buf[MAXSTR];
    /* Do nothing */

#ifdef DEBUG
    {
	TCHAR buf[MAXSTR];
	REDATA *prd = GlobalLock((HGLOBAL)hPort);
	if (prd == (REDATA *)NULL) {
	    SetLastError(ERROR_INVALID_HANDLE);
	    return FALSE;
	}
	if (prd->debug) {
	    wsprintf(buf, TEXT("REDMON SetPortTimeOuts: returning TRUE\r\n\
	    values = %d %d %d %d %d\r\n"),
		lpCTO->ReadIntervalTimeout,
		lpCTO->ReadTotalTimeoutMultiplier,
		lpCTO->ReadTotalTimeoutConstant,
		lpCTO->WriteTotalTimeoutMultiplier,
		lpCTO->WriteTotalTimeoutConstant);
	    write_string_to_log(prd, buf);
	}
	GlobalUnlock((HGLOBAL)hPort);
    }
#endif

    return TRUE;
}
 

/* Exported functions */


#ifdef UNICODE
#ifdef NT35
/* Windows NT 3.51 */
BOOL WINAPI _export
InitializeMonitor(LPWSTR pRegisterRoot)
{
    TCHAR buf[MAXSTR];
    LPTSTR p;
    if (lstrlen(pRegisterRoot) + 1 > sizeof(rekey) / sizeof(TCHAR)) 
	return FALSE;

    lstrcpy(rekey, pRegisterRoot);

    /* get help file name */
    GetModuleFileName(hdll, helpfile, sizeof(helpfile)/sizeof(TCHAR));
    p = helpfile + lstrlen(helpfile) - 1;
    while (p >= helpfile) {
	if (*p == '\\') {
	    p++;
	    break;
        }
	p--;
    }
    LoadString(hdll, IDS_HELPFILE, p, 
	sizeof(helpfile)/sizeof(TCHAR) - (int)(p-helpfile));

    return TRUE;
}

#else
/* Windows NT4 version */
MONITOREX mex;

LPMONITOREX WINAPI _export
InitializePrintMonitor(LPWSTR pRegisterRoot)
{
    TCHAR buf[MAXSTR];
    LPTSTR p;
#ifdef UNUSED
    wsprintf(buf, TEXT("InitializePrintMonitor: registrykey=%s"), 
		pRegisterRoot);
    MessageBox(NULL, buf, MONITORNAME, MB_OK);
#endif

    if (lstrlen(pRegisterRoot) + 1 > sizeof(rekey) / sizeof(TCHAR)) 
	return FALSE;

    lstrcpy(rekey, pRegisterRoot);

    mex.dwMonitorSize = sizeof(MONITOR);
    mex.Monitor.pfnEnumPorts = rEnumPorts;
    mex.Monitor.pfnOpenPort = rOpenPort;
    mex.Monitor.pfnOpenPortEx = NULL;
    mex.Monitor.pfnStartDocPort = rStartDocPort;
    mex.Monitor.pfnWritePort = rWritePort;
    mex.Monitor.pfnReadPort = rReadPort;
    mex.Monitor.pfnEndDocPort = rEndDocPort;
    mex.Monitor.pfnClosePort = rClosePort;
    mex.Monitor.pfnAddPort = rAddPort;
    mex.Monitor.pfnAddPortEx = rAddPortEx;
    mex.Monitor.pfnConfigurePort = rConfigurePort;
    mex.Monitor.pfnDeletePort = rDeletePort;
    mex.Monitor.pfnGetPrinterDataFromPort = NULL;
    mex.Monitor.pfnSetPortTimeOuts = rSetPortTimeOuts;

    /* get help file name */
    GetModuleFileName(hdll, helpfile, sizeof(helpfile)/sizeof(TCHAR));
    p = helpfile + lstrlen(helpfile) - 1;
    while (p >= helpfile) {
	if (*p == '\\') {
	    p++;
	    break;
        }
	p--;
    }
    LoadString(hdll, IDS_HELPFILE, p, 
	sizeof(helpfile)/sizeof(TCHAR) - (int)(p-helpfile));

    return &mex;
}
#endif	/* !NT35 */
#else
/* Windows 95 version */
BOOL WINAPI 
InitializeMonitorEx(LPTSTR pRegisterRoot, LPMONITOR pMonitor)
{
    TCHAR buf[MAXSTR];
    LPTSTR p;
#ifdef UNUSED
    wsprintf(buf, TEXT("InitializeMonitorEx: registrykey=%s pmonitor=%d"), 
		pRegisterRoot, pMonitor);
    MessageBox(NULL, buf, MONITORNAME, MB_OK);
#endif

    if (lstrlen(pRegisterRoot) + 1 > sizeof(rekey) / sizeof(TCHAR)) 
	return FALSE;

    lstrcpy(rekey, pRegisterRoot);

    pMonitor->pfnEnumPorts = rEnumPorts;
    pMonitor->pfnOpenPort = rOpenPort;
    pMonitor->pfnOpenPortEx = NULL;
    pMonitor->pfnStartDocPort = rStartDocPort;
    pMonitor->pfnWritePort = rWritePort;
    pMonitor->pfnReadPort = rReadPort;
    pMonitor->pfnEndDocPort = rEndDocPort;
    pMonitor->pfnClosePort = rClosePort;
    pMonitor->pfnAddPort = rAddPort;
    pMonitor->pfnConfigurePort = rConfigurePort;
    pMonitor->pfnDeletePort = rDeletePort;
    pMonitor->pfnGetPrinterDataFromPort = NULL;
    pMonitor->pfnSetPortTimeOuts = rSetPortTimeOuts;

    /* get help file name */
    GetModuleFileName(hdll, helpfile, sizeof(helpfile)/sizeof(TCHAR));
    p = helpfile + lstrlen(helpfile) - 1;
    while (p >= helpfile) {
	if (*p == '\\') {
	    p++;
	    break;
        }
	p--;
    }
    LoadString(hdll, IDS_HELPFILE, p, 
	sizeof(helpfile)/sizeof(TCHAR) - (int)(p-helpfile));

    return TRUE;

}
#endif


/* DLL entry point for Borland C++ */
#ifdef __BORLANDC__
#pragma argsused
#endif
BOOL WINAPI _export
DllEntryPoint(HINSTANCE hInst, DWORD fdwReason, LPVOID lpReserved)
{
    hdll = hInst;
    return TRUE;
}

#ifdef __BORLANDC__
#pragma argsused
#endif
/* DLL entry point for Microsoft Visual C++ */
BOOL WINAPI
DllMain(HINSTANCE hInst, DWORD fdwReason, LPVOID lpReserved)
{
    return DllEntryPoint(hInst, fdwReason, lpReserved);
}

BOOL
browse(HWND hwnd, UINT control, UINT filter, BOOL save)
{
OPENFILENAME ofn;
TCHAR szFilename[MAXSTR];	/* filename for OFN */
TCHAR szFilter[MAXSTR];
TCHAR szDir[MAXSTR];
TCHAR cReplace;
int i;

BOOL flag;
	/* setup OPENFILENAME struct */
	FillMemory((PVOID)&ofn, sizeof(ofn), 0);
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = hwnd;
	ofn.lpstrFile = szFilename;
	ofn.nMaxFile = sizeof(szFilename);
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

	GetDlgItemText(hwnd, control, szFilename, 
		sizeof(szFilename)/sizeof(TCHAR) - 1);

	if ((lstrlen(szFilename) == 2) && (szFilename[1] == ':'))
	    lstrcat(szFilename, BACKSLASH);

	if (lstrlen(szFilename)) {
	    lstrcpy(szDir, szFilename);
	    for (i=lstrlen(szDir)-1; i; i--) {
		if (szDir[i] == '\\') {
		    lstrcpy(szFilename, szDir+i+1);
		    szDir[i+1] = '\0';
		    ofn.lpstrInitialDir = szDir;
		    break;
		}
	    }
	}

	if (LoadString(hdll, filter, szFilter, 
	    sizeof(szFilter)/sizeof(TCHAR) -1 )) {
	    cReplace = szFilter[lstrlen(szFilter)-1];
	    for (i=0; szFilter[i] != '\0'; i++)
	        if (szFilter[i] == cReplace)
		    szFilter[i] = '\0';
	    ofn.lpstrFilter = szFilter;
	    ofn.nFilterIndex = 0;
	}

	/* call the common dialog box */
	if (save)
	    flag = GetSaveFileName(&ofn);
	else
	    flag = GetOpenFileName(&ofn);

	if (flag)
	    SetDlgItemText(hwnd, control, szFilename);
	return flag;
}

int
lstrcmpn(LPTSTR s1, LPTSTR s2, int len2)
{
int flag = 0;
    for (; len2 && *s1 && *s2 ; len2--) {
	flag = *s1 - *s2;
	s1++;
	s2++;
    }
    return flag;
}

/* Get the directory for temporary files */
/* Store the result in buf */
/* Don't copy if length is greater than len characters */
BOOL 
get_temp(LPTSTR value, int len) 
{
TCHAR buf[256];
DWORD dwLength;
    /* If we run from the Windows NT SYSTEM account, many 
     * environment variables aren't set.  We need to look
     * in several places to find the name of a directory
     * for temporary files.
     */

    if (!len)
	return FALSE;
    value[0] = '\0';

    dwLength = GetEnvironmentVariable(TEXT("TEMP"), value, len);
    if (dwLength == 0)
        dwLength = GetEnvironmentVariable(TEXT("TMP"), value, len);

    if (dwLength == 0) {
	HKEY hkey;
	    DWORD cbData, keytype;
        if ( RegOpenKeyEx(HKEY_USERS, TEXT(".DEFAULT\\Environment"), 0, 
		KEY_ALL_ACCESS, &hkey) == ERROR_SUCCESS ) {
	    keytype = REG_SZ;
	    value[0] = '\0';
	    cbData = sizeof(buf);
	    if (RegQueryValueEx(hkey, TEXT("TEMP"), 0, &keytype, 
		(LPBYTE)buf, &cbData) == ERROR_SUCCESS)
		dwLength = cbData;
	    if (dwLength == 0) {
		buf[0] = '\0';
		cbData = sizeof(buf);
		if (RegQueryValueEx(hkey, TEXT("TMP"), 0, &keytype, 
		    (LPBYTE)buf, &cbData) == ERROR_SUCCESS)
		    dwLength = cbData;
	    }
	    RegCloseKey(hkey);
	    if (dwLength) {
		LPTSTR systemdrive = TEXT("%SystemDrive%");
	        if (lstrcmpn(buf, systemdrive, 
			lstrlen(systemdrive)) == 0) {
		    LPTSTR p = buf + lstrlen(systemdrive);
		    if (GetEnvironmentVariable(TEXT("SystemDrive"),
			buf, sizeof(buf)/sizeof(TCHAR))) {
			if (lstrlen(buf) + lstrlen(p) < len - 1) {
			    lstrcpy(value, buf);
			    lstrcat(value, p);
			    dwLength = lstrlen(value);
			}
			else
			    dwLength = 0;
		    }
		    else
		        dwLength = 0;
	        }
	        else {
		    lstrcpy(value, buf);
		}
	    }
	}
    }

    if (dwLength == 0)
	GetWindowsDirectory(value, len);
    if (dwLength)
	return TRUE;
    return FALSE;
}

/* create a temporary file.
 * If temporary filename is shorter than len, create the file
 * and store name in filename.
 * Return TRUE if success.
 */
int
create_tempfile(LPTSTR filename, DWORD len)
{
TCHAR temp[256];
TCHAR buf[256];
LPTSTR p;
int i;
HANDLE hf;

    get_temp(temp, sizeof(temp)/sizeof(TCHAR));

    if (lstrlen(temp) > 0) {
	p = temp + lstrlen(temp) - 1;
	if (*p != '\\')
	    lstrcat(temp, TEXT("\\"));
    }
    lstrcat(temp, TEXT("RedMon"));
    for (i=0; i<100000; i++) {
	wsprintf(buf, TEXT("%s%02d.%03d"), temp, i / 1000, i % 1000);
	if (len < lstrlen(buf) + 1)
	    return FALSE;
        hf = CreateFile(buf, GENERIC_WRITE, 0 /* no sharing */, 
	    NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hf != INVALID_HANDLE_VALUE) {
	    CloseHandle(hf);
	    lstrcpy(filename, buf);
	    return TRUE;
	}
    }
    return FALSE;
}



/* True Win32 method, using OpenPrinter, WritePrinter etc. */
int 
redmon_printfile(REDATA * prd, TCHAR *filename)
{
HGLOBAL hbuffer;
BYTE *buffer;
DWORD cbRead;
HANDLE hread;

    if (prd->printername[0] == '\0')
	return FALSE;

    /* allocate buffer for reading data */
    if ((hbuffer = GlobalAlloc(GPTR, (DWORD)PRINT_BUF_SIZE)) == NULL)
	return FALSE;
    if ((buffer = (BYTE *)GlobalLock(hbuffer)) == (BYTE *)NULL) {
	GlobalFree(hbuffer);
        return FALSE;
    }
	
    /* open file to print */
    if ((hread = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, 
	NULL, OPEN_EXISTING, 0, NULL)) == INVALID_HANDLE_VALUE) {
	if (prd->debug) {
	    DWORD err = GetLastError();
	    TCHAR buf[256];
	    wsprintf(buf, TEXT("CreateFile() failed, error code = %d\r\n"), 
		err);
	    write_string_to_log(prd, buf);
	    write_error(prd, err);
	}
	GlobalUnlock(hbuffer);
	GlobalFree(hbuffer);
	return FALSE;
    }

    /* open a printer */
    if (!redmon_open_printer(prd)) {
	CloseHandle(hread);
	GlobalUnlock(hbuffer);
	GlobalFree(hbuffer);
	return FALSE;
    }

    while (ReadFile(hread, buffer, PRINT_BUF_SIZE, &cbRead, NULL)
		&& cbRead) {
	if (!redmon_write_printer(prd, buffer, cbRead)) {
	    CloseHandle(hread);
	    redmon_abort_printer(prd);
	    return FALSE;
	}
    }
    CloseHandle(hread);
    GlobalUnlock(hbuffer);
    GlobalFree(hbuffer);

    redmon_close_printer(prd);

    return TRUE;
}



BOOL
redmon_open_printer(REDATA *prd)
{
    TCHAR buf[256];
    DOC_INFO_1 di;
    PRINTER_INFO_2 *prinfo;
    HGLOBAL hglobal;
    DWORD needed, count;
    BOOL fail = FALSE;

    prd->printer = INVALID_HANDLE_VALUE;
    if (lstrlen(prd->printername) == 0)
	return FALSE;

    /* open a printer */
    if (!OpenPrinter(prd->printername, &prd->printer, NULL)) {
	if (prd->debug) {
	    DWORD err = GetLastError();
	    wsprintf(buf, 
	      TEXT("OpenPrinter() failed for \042%s\042, error code = %d\r\n"), 
		prd->printername, err);
	    write_string_to_log(prd, buf);
	    write_error(prd, err);
	}
	return FALSE;
    }
    /* from here until ClosePrinter, should AbortPrinter on error */

    /* make sure we don't try to write to our own port! */
    needed = 0;
    GetPrinter(prd->printer, 2, NULL, 0, &needed);
    if (needed == 0)
	fail = TRUE;
    if (!fail) 
        if ((hglobal = GlobalAlloc(GPTR, (DWORD)needed)) == NULL)
	    fail = TRUE;
    if (!fail) 
	if ((prinfo = (PRINTER_INFO_2 *)GlobalLock(hglobal)) == 
		(PRINTER_INFO_2 *)NULL) {
	    GlobalFree(hglobal);
	    fail = TRUE;
	}
    if (!fail) {
	count = needed;
	if (!GetPrinter(prd->printer, 2, (LPBYTE)prinfo, count, &needed)) {
	    GlobalUnlock(hglobal);
	    GlobalFree(hglobal);
	    fail = TRUE;
	}
    }
    if (!fail) {
	if (lstrcmp(prinfo->pPortName, prd->portname) == 0) {
	  write_string_to_log(prd,
	    TEXT("Talking to yourself is a sign of madness.\r\n"));
	  write_string_to_log(prd,
	    TEXT("Writing to the port being redirected is not permitted.\r\n"));
	    fail = TRUE;
	}
	GlobalUnlock(hglobal);
	GlobalFree(hglobal);
    }
    if (fail) {
	redmon_abort_printer(prd);
	return FALSE;
    }


    di.pDocName = prd->DocName;	/* keep original document name */
    di.pOutputFile = NULL;
    di.pDatatype = TEXT("RAW");	/* for available types see EnumPrintProcessorDatatypes */
    if (!StartDocPrinter(prd->printer, 1, (LPBYTE)&di)) {
	if (prd->debug) {
	    DWORD err = GetLastError();
	    wsprintf(buf, 
		TEXT("StartDocPrinter() failed, error code = %d\r\n"), 
		err);
	    write_string_to_log(prd, buf);
	    write_error(prd, err);
	}
	redmon_abort_printer(prd);
	return FALSE;
    }
    return TRUE;
}

BOOL
redmon_abort_printer(REDATA *prd)
{
    if (prd->debug)
	write_string_to_log(prd, TEXT("AbortPrinter\r\n"));
    if (prd->printer != INVALID_HANDLE_VALUE)
	AbortPrinter(prd->printer);
    prd->printer = INVALID_HANDLE_VALUE;
    return TRUE;
}

BOOL
redmon_close_printer(REDATA *prd)
{
    TCHAR buf[256];
    if (prd->printer == INVALID_HANDLE_VALUE)
	return TRUE;
    if (!EndDocPrinter(prd->printer)) {
	if (prd->debug) {
	    DWORD err = GetLastError();
	    wsprintf(buf, 
		TEXT("EndDocPrinter() failed, error code = %d\r\n"), err);
	    write_string_to_log(prd, buf);
	    write_error(prd, err);
	}
	redmon_abort_printer(prd);
	return FALSE;
    }

    if (!ClosePrinter(prd->printer)) {
	if (prd->debug) {
	    DWORD err = GetLastError();
	    wsprintf(buf, 
		TEXT("ClosePrinter() failed, error code = %d\r\n"), 
		err);
	    write_string_to_log(prd, buf);
	    write_error(prd, err);
	}
        prd->printer = INVALID_HANDLE_VALUE;
	return FALSE;
    }
    prd->printer = INVALID_HANDLE_VALUE;
    return TRUE;
}

BOOL
redmon_write_printer(REDATA *prd, BYTE *ptr, DWORD len)
{
DWORD cbWritten;
    if (prd->printer == INVALID_HANDLE_VALUE)
	return FALSE;
    if (!WritePrinter(prd->printer, (LPVOID)ptr, len, &cbWritten)) {
	TCHAR buf[256];
	DWORD err = GetLastError();
	wsprintf(buf, 
	    TEXT("WritePrinter() failed, error code = %d\r\n"), err);
	write_string_to_log(prd, buf);
	write_error(prd, err);
	return FALSE;
    }
    if (len != cbWritten) {
	TCHAR buf[256];
	wsprintf(buf, 
	    TEXT("WritePrinter() error: wrote %d bytes of %d\r\n"), 
	    cbWritten, len);
	write_string_to_log(prd, buf);
    }
    prd->printer_bytes += cbWritten;
    return TRUE;
}


BOOL CALLBACK
AbortProc(HDC hdcPrn, int code)
{
    if (code == SP_OUTOFDISK)
	return FALSE;	/* cancel job */
    return TRUE;
}

/* This may be called if an error has occurred,
 * to print out the log file as an error report.
 */
BOOL 
redmon_print_error(REDATA *prd)
{
    HDC hdc;
    TCHAR driverbuf[2048];
    TCHAR *device, *driver, *output, *p;
    DOCINFO di;
    HFONT hfont;
    HANDLE hfile;
    LOGFONT lf;
    CHAR line[128];
    int xdpi, ydpi;
    int x, y, ymin, ymax, yskip;
    BOOL error = FALSE;
    

    /* open printer */
    device = prd->printername;
    GetProfileString(TEXT("Devices"), device, TEXT(""), driverbuf, 
	sizeof(driverbuf) / sizeof(TCHAR));
    p = driverbuf;
    driver = p;
    while (*p && *p != ',')
	p++;
    if (*p)
	*p++ = '\0';
    output = p;
    while (*p && *p != ',')
	p++;
    if (*p)
	*p++ = '\0';
 
    if ( (hdc = CreateDC(driver, device, NULL, NULL)) == NULL) {
	return FALSE;
    }

    SetAbortProc(hdc, AbortProc);
    di.cbSize = sizeof(di);
    di.lpszDocName = TEXT("RedMon error report");
    di.lpszOutput = output;
    if ((error = (StartDoc(hdc, &di) == SP_ERROR)) != 0) {
	TCHAR buf[1024];
	DWORD last_error = GetLastError();
	wsprintf(buf, 
	    TEXT("StartDoc \042%s\042,\042%s\042,\042%s\042 Error=%d"),
	    device, di.lpszDocName, di.lpszOutput, last_error);
	MessageBox(HWND_DESKTOP, buf, TEXT("RedMon"), MB_OK);
    }
	
    if (!error && (StartPage(hdc) != SP_ERROR)) {
	xdpi = GetDeviceCaps(hdc, LOGPIXELSX);
	ydpi = GetDeviceCaps(hdc, LOGPIXELSY);
	memset(&lf, 0, sizeof(LOGFONT));
	lf.lfHeight = 10 * ydpi / 72;	/* 10 pts high */
	lstrcpy(lf.lfFaceName, TEXT("Arial"));
	hfont = CreateFontIndirect(&lf);
	SelectObject(hdc, hfont);

	ymin = 36 * ydpi / 72;	/* top of page */
	ymax = 756 * ydpi / 72;	/* bottom of page */
	yskip = 10 * ydpi / 72;	/* line spacing */
	x = 36 * xdpi / 72;		/* left margin */
	y = ymin;

	/* write RedMon header */
	p = TEXT("RedMon error report");
	TextOut(hdc, x, y, p, lstrlen(p));
	y+=yskip;
	lstrcpy(driverbuf, copyright);
	p = driverbuf + lstrlen(driverbuf) - 1;
	if (*p == '\n')
	    *p = '\0';
	p = driverbuf;
	TextOut(hdc, x, y, p, lstrlen(p));
	y+=yskip;
	lstrcpy(driverbuf, version);
	p = driverbuf + lstrlen(driverbuf) - 1;
	if (*p == '\n')
	    *p = '\0';
	p = driverbuf;
	TextOut(hdc, x, y, p, lstrlen(p));
	y+=yskip;
	p = driverbuf;
	wsprintf(driverbuf, TEXT("  Port=\042%s\042  Printer=\042%s\042"), 
	    prd->portname, prd->pPrinterName);
	TextOut(hdc, x, y, p, lstrlen(p));
	y+=yskip;
	wsprintf(driverbuf, TEXT("  DocumentName=\042%s\042"), prd->DocName);
	TextOut(hdc, x, y, p, lstrlen(p));
	y+=yskip;
/*
	wsprintf(driverbuf, TEXT("  Command=\042%s\042), prd->command);
	TextOut(hdc, x, y, p, lstrlen(p));
	y+=yskip;
*/
	y+=yskip;


	if (prd->logname[0] != '\0') {
	    /* log file was written (but isn't currently open) */
	    /* open log file */
	    if ((hfile = CreateFile(prd->logname, GENERIC_READ, 
		FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)) 
		!= INVALID_HANDLE_VALUE) {
		/* copy log file to printer */
		int i = 0;
		DWORD dwRead = 0;
		BOOL eol = FALSE; 
		while (ReadFile(hfile, &line[i], 1, &dwRead, NULL) &&
		    (dwRead == 1)) {
		    eol = FALSE;
		    switch (line[i]) {
			case '\r':
			    /* do nothing */
			    break;
			case '\n':
			    /* end of line */
			    eol = TRUE;
			    line[i]='\0';
			    break;
			default:
			    i++;
			    if (i > 80) {
				eol = TRUE;
				line[i]='\0';
			    }
		    }

		    if (eol) {
			if (y >= ymax) {
			    EndPage(hdc);
			    StartPage(hdc);
			    y = ymin;
			}
			TextOutA(hdc, x, y, line, i);
			y+=yskip;
			i = 0;
		    }
		    dwRead = 0;
		}
		if (i)
		    TextOutA(hdc, x, y, line, i);
		CloseHandle(hfile);
	    }
	}
	/* close printer */
	EndPage(hdc);
	EndDoc(hdc);
    }

    DeleteDC(hdc);
}


/* Add Port dialog box */
BOOL CALLBACK 
AddDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message) {
        case WM_INITDIALOG:
	    {
	    TCHAR buf[MAXSTR];
	    ADDPARAM *ap;

	    /* save pointer to port name */
	    SetWindowLong(hDlg, GWL_USERDATA, (LONG)lParam);
	    
	    ap = (ADDPARAM *)lParam;

	    SetDlgItemText(hDlg, IDC_PORTNAME, ap->portname);
	    SetDlgItemText(hDlg, IDC_PORTDESC, ap->portdesc);
	    SetForegroundWindow(hDlg);
	    }
            return( TRUE);
        case WM_COMMAND:
            switch(LOWORD(wParam)) {
	      case IDC_HELPBUTTON:
		  {TCHAR buf[MAXSTR];
		  LoadString(hdll, IDS_HELPADD, buf, 
			sizeof(buf)/sizeof(TCHAR) - 1);
		  WinHelp(hDlg, helpfile, HELP_KEY, (DWORD)buf);
		  }
                  return(TRUE);
              case IDC_PORTNAME:
              case IDC_PORTDESC:
                  return(TRUE);
	      case IDOK:
		{ ADDPARAM *ap = (ADDPARAM *)GetWindowLong(hDlg, GWL_USERDATA);
		  GetDlgItemText(hDlg, IDC_PORTNAME, ap->portname, 
			MAXSHORTSTR-1);
		  GetDlgItemText(hDlg, IDC_PORTDESC, ap->portdesc, 
			MAXSHORTSTR-1);
		}
		EndDialog(hDlg, TRUE);
		return(TRUE);
	      case IDCANCEL:
                EndDialog(hDlg, FALSE);
                return(TRUE);
              default:
                return(FALSE);
            }
        default:
            return(FALSE);
    }
}

/* Log file dialog box */
BOOL CALLBACK 
LogfileDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message) {
        case WM_INITDIALOG:
	    {
	    HKEY hkey;
	    TCHAR buf[MAXSTR];
	    TCHAR str[MAXSTR];
	    DWORD cbData, keytype;
	    DWORD dwtemp;

	    /* save pointer to port name */
	    SetWindowLong(hDlg, GWL_USERDATA, (LONG)lParam);

	    lstrcpy(buf, rekey);
	    lstrcat(buf, BACKSLASH);
	    lstrcat(buf, PORTSNAME);
	    lstrcat(buf, BACKSLASH);
	    lstrcat(buf, (LPTSTR)lParam);
	    if ( RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, 
		    KEY_ALL_ACCESS, &hkey) != ERROR_SUCCESS ) {
		/* we can't do anything */
		LoadString(hdll, IDS_CONFIGUNKNOWN, buf, 
		    sizeof(buf)/sizeof(TCHAR) - 1);
		SetWindowText(hDlg, buf);
		return TRUE;
	    }

	    keytype = REG_SZ;
	    buf[0] = '\0';
	    cbData = sizeof(buf);
	    RegQueryValueEx(hkey, LOGNAMEKEY, 0, &keytype, 
		(LPBYTE)buf, &cbData);
	    SetDlgItemText(hDlg, IDC_LOGNAME, buf);

#ifdef BETA
	    dwtemp = 1;	/* beta versions default to debug enabled */
#else
	    dwtemp = 0;
#endif
	    keytype = REG_DWORD;
	    cbData = sizeof(dwtemp);
	    RegQueryValueEx(hkey, LOGDEBUGKEY, 0, &keytype, 
		(LPBYTE)&dwtemp, &cbData);
	    SendDlgItemMessage(hDlg, IDC_LOGDEBUG, BM_SETCHECK, dwtemp, 0);

	    dwtemp = 0;
	    keytype = REG_DWORD;
	    cbData = sizeof(dwtemp);
	    RegQueryValueEx(hkey, LOGUSEKEY, 0, &keytype, 
		(LPBYTE)&dwtemp, &cbData);

	    RegCloseKey(hkey);
	    
	    SendDlgItemMessage(hDlg, IDC_LOGUSE, BM_SETCHECK, dwtemp, 0);
	    EnableWindow(GetDlgItem(hDlg, IDC_LOGNAMEPROMPT), dwtemp);
	    EnableWindow(GetDlgItem(hDlg, IDC_LOGNAME), dwtemp);
	    EnableWindow(GetDlgItem(hDlg, IDC_LOGDEBUG), dwtemp);
	    EnableWindow(GetDlgItem(hDlg, IDC_BROWSE), dwtemp);

	    LoadString(hdll, IDS_CONFIGLOGFILE, str, 
		sizeof(str)/sizeof(TCHAR) - 1);
	    wsprintf(buf, str, (LPSTR)lParam);
	    SetWindowText(hDlg, buf);
	    }
            return( TRUE);
        case WM_COMMAND:
            switch(LOWORD(wParam)) {
	      case IDC_HELPBUTTON:
		  {TCHAR buf[MAXSTR];
		  LoadString(hdll, IDS_HELPLOG, buf, 
			sizeof(buf)/sizeof(TCHAR) - 1);
		  WinHelp(hDlg, helpfile, HELP_KEY, (DWORD)buf);
		  }
                  return(TRUE);
	      case IDC_BROWSE:
		  browse(hDlg, IDC_LOGNAME, IDS_FILTER_TXT, TRUE);
                  return(TRUE);
	      case IDC_LOGUSE:
		  if (HIWORD(wParam) == BN_CLICKED) {
		    BOOL enabled = (BOOL)SendDlgItemMessage(hDlg, 
				IDC_LOGUSE, BM_GETCHECK, 0, 0);
		    enabled = !enabled;
		    SendDlgItemMessage(hDlg, IDC_LOGUSE, 
			BM_SETCHECK, enabled, 0);
		    EnableWindow(GetDlgItem(hDlg, IDC_LOGNAMEPROMPT), enabled);
		    EnableWindow(GetDlgItem(hDlg, IDC_LOGNAME), enabled);
		    EnableWindow(GetDlgItem(hDlg, IDC_LOGDEBUG), enabled);
		    EnableWindow(GetDlgItem(hDlg, IDC_BROWSE), enabled);
		   }
		   return TRUE;
	      case IDC_LOGFILE:
                    return(TRUE);
/* should we fall through to IDOK */
	      case IDOK:
		{
		HKEY hkey;
		TCHAR buf[MAXSTR];
		BOOL enabled;
		DWORD dwtemp;
		BOOL success;
		lstrcpy(buf, rekey);
		lstrcat(buf, BACKSLASH);
		lstrcat(buf, PORTSNAME);
		lstrcat(buf, BACKSLASH);
		lstrcat(buf, (LPTSTR)GetWindowLong(hDlg, GWL_USERDATA));
		if ( RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, 
			KEY_ALL_ACCESS, &hkey) == ERROR_SUCCESS ) {
		    enabled = (BOOL)SendDlgItemMessage(hDlg, IDC_LOGUSE,
			BM_GETCHECK, 0, 0);
		    dwtemp = (DWORD)enabled;
		    RegSetValueEx(hkey, LOGUSEKEY, 0, REG_DWORD,
			(CONST BYTE *)&dwtemp, sizeof(DWORD));
		    if (enabled) {
			/* get log filename */
		        GetDlgItemText(hDlg, IDC_LOGNAME, buf, 
			    sizeof(buf)/sizeof(TCHAR) - 1);
		        RegSetValueEx(hkey, LOGNAMEKEY, 0, REG_SZ,
			    (CONST BYTE *)buf, (lstrlen(buf)+1)*sizeof(TCHAR));
			/* get debug flag */
		        dwtemp = SendDlgItemMessage(hDlg, IDC_LOGDEBUG,
			    BM_GETCHECK, 0, 0);
			RegSetValueEx(hkey, LOGDEBUGKEY, 0, REG_DWORD,
			    (CONST BYTE *)&dwtemp, sizeof(DWORD));
		    }
		    RegCloseKey(hkey);
		}
		}
		EndDialog(hDlg, TRUE);
		return(TRUE);
	      case IDCANCEL:
                EndDialog(hDlg, FALSE);
                return(TRUE);
              default:
                return(FALSE);
            }
        default:
            return(FALSE);
    }
}

/* setup up combo box with list of printers */
void
init_printer_list(HWND hDlg, LPTSTR portname)
{
int i;
DWORD count, needed;
PRINTER_INFO_2 *prinfo;
LPTSTR enumbuffer;
HGLOBAL hglobal;
BOOL fail = FALSE;

    /* enumerate all available printers */
    needed = 0;
    EnumPrinters(PRINTER_ENUM_CONNECTIONS | PRINTER_ENUM_LOCAL, 
	NULL, 2, NULL, 0, &needed, &count);
    if (needed == 0)
	fail = TRUE;
    if (!fail) 
        if ((hglobal = GlobalAlloc(GPTR, (DWORD)needed)) == NULL)
	    fail = TRUE;
    if (!fail) 
	if ((enumbuffer = (LPTSTR)GlobalLock(hglobal)) == (LPTSTR)NULL) {
	    GlobalFree(hglobal);
	    fail = TRUE;
	}

    if (!fail)
	if (!EnumPrinters(PRINTER_ENUM_CONNECTIONS | PRINTER_ENUM_LOCAL, 
	    NULL, 2, (LPBYTE)enumbuffer, needed, &needed, &count)) {
	    GlobalUnlock(hglobal);
	    GlobalFree(hglobal);
	    fail = TRUE;
	}

    if (fail) {
	EnableWindow(GetDlgItem(hDlg, IDC_PRINTER), FALSE);
    }
    else {
	prinfo = (PRINTER_INFO_2 *)enumbuffer;
	SendDlgItemMessage(hDlg, IDC_PRINTER, CB_RESETCONTENT, 
	    (WPARAM)0, (LPARAM)0);
	for (i=0; i<count; i++) {
	    if (lstrcmp(prinfo[i].pPortName, portname) != 0)
	      SendDlgItemMessage(hDlg, IDC_PRINTER, CB_ADDSTRING, 0, 
		(LPARAM)(prinfo[i].pPrinterName));
	}
    }
}


/* Configure Port dialog box */
BOOL CALLBACK 
ConfigDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch(message) {
        case WM_INITDIALOG:
	    {
	    HKEY hkey;
	    TCHAR buf[MAXSTR];
	    TCHAR str[MAXSTR];
	    DWORD cbData, keytype;
	    DWORD show;
	    DWORD output;
	    DWORD runuser;
	    DWORD delay;
	    DWORD printerror;
	    int i;

	    /* save pointer to port name */
	    SetWindowLong(hDlg, GWL_USERDATA, (LONG)lParam);

	    lstrcpy(buf, rekey);
	    lstrcat(buf, BACKSLASH);
	    lstrcat(buf, PORTSNAME);
	    lstrcat(buf, BACKSLASH);
	    lstrcat(buf, (LPTSTR)lParam);
	    if ( RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, 
		    KEY_ALL_ACCESS, &hkey) != ERROR_SUCCESS ) {
		/* we can't do anything */
		LoadString(hdll, IDS_CONFIGUNKNOWN, buf, 
		    sizeof(buf)/sizeof(TCHAR) - 1);
		SetWindowText(hDlg, buf);
		return TRUE;
	    }

	    keytype = REG_SZ;
	    buf[0] = '\0';
	    cbData = sizeof(buf);
	    RegQueryValueEx(hkey, COMMANDKEY, 0, &keytype, 
		(LPBYTE)buf, &cbData);
	    SetDlgItemText(hDlg, IDC_COMMAND, buf);
	    buf[0] = '\0';
	    cbData = sizeof(buf);
	    RegQueryValueEx(hkey, ARGKEY, 0, &keytype, 
		(LPBYTE)buf, &cbData);
	    SetDlgItemText(hDlg, IDC_ARGS, buf);

	    cbData = sizeof(DWORD);
	    keytype = REG_DWORD;
	    output = 0;
	    RegQueryValueEx(hkey, OUTPUTKEY, 0, &keytype, 
		(LPBYTE)&output, &cbData);
	    SendDlgItemMessage(hDlg, IDC_OUTPUT, CB_RESETCONTENT,
		(WPARAM)0, (LPARAM)0);
	    for (i=0; i<=OUTPUT_HANDLE; i++) {
		LoadString(hdll, IDS_OUTPUTBASE + i, buf, 
		    sizeof(buf)/sizeof(TCHAR) - 1);
		SendDlgItemMessage(hDlg, IDC_OUTPUT, CB_ADDSTRING,
		    (WPARAM)0, (LPARAM)buf);
	    }
	    SendDlgItemMessage(hDlg, IDC_OUTPUT, CB_SETCURSEL,
		(WPARAM)output, (LPARAM)0);

	    init_printer_list(hDlg, (LPTSTR)lParam);
	    keytype = REG_SZ;
	    buf[0] = '\0';
	    cbData = sizeof(buf);
	    RegQueryValueEx(hkey, PRINTERKEY, 0, &keytype, 
		(LPBYTE)buf, &cbData);
	    if (SendDlgItemMessage(hDlg, IDC_PRINTER, CB_SELECTSTRING, -1, 
	        (LPARAM)(buf)) == CB_ERR)
	        SendDlgItemMessage(hDlg, IDC_PRINTER, CB_SETCURSEL, 0, 0);

	    EnableWindow(GetDlgItem(hDlg, IDC_PRINTER), 
		output != OUTPUT_SELF);
	    EnableWindow(GetDlgItem(hDlg, IDC_PRINTERTEXT), 
		output != OUTPUT_SELF);

	    cbData = sizeof(DWORD);
	    keytype = REG_DWORD;
	    show = 0;
	    RegQueryValueEx(hkey, SHOWKEY, 0, &keytype, 
		(LPBYTE)&show, &cbData);
	    SendDlgItemMessage(hDlg, IDC_SHOW, CB_RESETCONTENT,
		(WPARAM)0, (LPARAM)0);
	    for (i=0; i<=SHOW_HIDE; i++) {
		LoadString(hdll, IDS_SHOWBASE + i, buf, 
		    sizeof(buf)/sizeof(TCHAR) - 1);
		SendDlgItemMessage(hDlg, IDC_SHOW, CB_ADDSTRING,
		    (WPARAM)0, (LPARAM)buf);
	    }
	    SendDlgItemMessage(hDlg, IDC_SHOW, CB_SETCURSEL,
		(WPARAM)show, (LPARAM)0);

	    cbData = sizeof(DWORD);
	    keytype = REG_DWORD;
	    runuser = 0;
	    RegQueryValueEx(hkey, RUNUSERKEY, 0, &keytype, 
		(LPBYTE)&runuser, &cbData);
	    SendDlgItemMessage(hDlg, IDC_RUNUSER, BM_SETCHECK,
		(WPARAM)runuser, (LPARAM)0);
#if !defined(UNICODE) || defined(NT35) || defined(__BORLANDC__)
	    EnableWindow(GetDlgItem(hDlg, IDC_RUNUSER), FALSE);
#endif

	    delay = DEFAULT_DELAY;
	    keytype = REG_DWORD;
	    cbData = sizeof(delay);
	    RegQueryValueEx(hkey, DELAYKEY, 0, &keytype, 
		(LPBYTE)&delay, &cbData);
	    SetDlgItemInt(hDlg, IDC_DELAY, delay, FALSE);

	    keytype = REG_DWORD;
	    cbData = sizeof(printerror);
	    printerror = 0;
	    RegQueryValueEx(hkey, PRINTERRORKEY, 0, &keytype, 
		(LPBYTE)&printerror, &cbData);
	    SendDlgItemMessage(hDlg, IDC_PRINTERROR, BM_SETCHECK, printerror, 0);
	    RegCloseKey(hkey);

	    LoadString(hdll, IDS_CONFIGPROP, str, 
		sizeof(str)/sizeof(TCHAR) - 1);
	    wsprintf(buf, str, (LPSTR)lParam);
	    SetWindowText(hDlg, buf);
	    SetForegroundWindow(hDlg);
	    }
            return( TRUE);
        case WM_COMMAND:
            switch(LOWORD(wParam)) {
	      case IDC_HELPBUTTON:
		  {TCHAR buf[MAXSTR];
		  LoadString(hdll, IDS_HELPCONFIG, buf, 
			sizeof(buf)/sizeof(TCHAR) - 1);
		  WinHelp(hDlg, helpfile, HELP_KEY, (DWORD)buf);
		  }
                  return(TRUE);
	      case IDC_BROWSE:
		  browse(hDlg, IDC_COMMAND, IDS_FILTER_EXE, FALSE);
                  return(TRUE);
	      case IDC_LOGFILE:
		  DialogBoxParam(hdll, MAKEINTRESOURCE(IDD_CONFIGLOG), hDlg, 
			LogfileDlgProc, 
			(LPARAM)GetWindowLong(hDlg, GWL_USERDATA));
                  return(TRUE);
	      case IDC_OUTPUT:
		  if (HIWORD(wParam) == CBN_SELCHANGE)  {
		      BOOL enabled = (SendDlgItemMessage(hDlg, IDC_OUTPUT, 
			CB_GETCURSEL, (WPARAM)0, (LPARAM)0) != OUTPUT_SELF);
		      EnableWindow(GetDlgItem(hDlg, IDC_PRINTER), enabled); 
		      EnableWindow(GetDlgItem(hDlg, IDC_PRINTERTEXT), enabled); 
		      EnableWindow(GetDlgItem(hDlg, IDC_PRINTERROR), enabled); 
		  }
                  return(TRUE);
              case IDC_COMMAND:
              case IDC_ARGS:
                    return(TRUE);
/* should we fall through to IDOK */
	      case IDOK:
		{
		HKEY hkey;
		TCHAR buf[MAXSTR];
		DWORD show;
		DWORD output;
		DWORD runuser;
		DWORD delay;
		BOOL success;
		DWORD printerror;
		lstrcpy(buf, rekey);
		lstrcat(buf, BACKSLASH);
		lstrcat(buf, PORTSNAME);
		lstrcat(buf, BACKSLASH);
		lstrcat(buf, (LPTSTR)GetWindowLong(hDlg, GWL_USERDATA));
		if ( RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, 
			KEY_ALL_ACCESS, &hkey) == ERROR_SUCCESS ) {
		    GetDlgItemText(hDlg, IDC_COMMAND, buf, 
			sizeof(buf)/sizeof(TCHAR) - 1);
		    RegSetValueEx(hkey, COMMANDKEY, 0, REG_SZ,
			(CONST BYTE *)buf, (lstrlen(buf)+1)*sizeof(TCHAR));
		    GetDlgItemText(hDlg, IDC_ARGS, buf, 
			sizeof(buf)/sizeof(TCHAR) - 1);
		    RegSetValueEx(hkey, ARGKEY, 0, REG_SZ,
			(CONST BYTE *)buf, (lstrlen(buf)+1)*sizeof(TCHAR));
		    output = SendDlgItemMessage(hDlg, IDC_OUTPUT, CB_GETCURSEL,
			(WPARAM)0, (LPARAM)0);
		    RegSetValueEx(hkey, OUTPUTKEY, 0, REG_DWORD,
			(CONST BYTE *)&output, sizeof(DWORD));
		    GetDlgItemText(hDlg, IDC_PRINTER, buf, 
			sizeof(buf)/sizeof(TCHAR) - 1);
		    RegSetValueEx(hkey, PRINTERKEY, 0, REG_SZ,
			(CONST BYTE *)buf, (lstrlen(buf)+1)*sizeof(TCHAR));
		    show = SendDlgItemMessage(hDlg, IDC_SHOW, CB_GETCURSEL,
			(WPARAM)0, (LPARAM)0);
		    RegSetValueEx(hkey, SHOWKEY, 0, REG_DWORD,
			(CONST BYTE *)&show, sizeof(DWORD));
		    runuser = SendDlgItemMessage(hDlg, IDC_RUNUSER, BM_GETCHECK,
			(WPARAM)0, (LPARAM)0);
		    RegSetValueEx(hkey, RUNUSERKEY, 0, REG_DWORD,
			(CONST BYTE *)&runuser, sizeof(DWORD));
		    delay = (DWORD)GetDlgItemInt(hDlg, IDC_DELAY, 
				&success, FALSE);
		    if (success)
		        RegSetValueEx(hkey, DELAYKEY, 0, REG_DWORD,
			    (CONST BYTE *)&delay, sizeof(DWORD));
		    if (output != OUTPUT_SELF) {
			printerror = SendDlgItemMessage(hDlg, IDC_PRINTERROR,
			    BM_GETCHECK, 0, 0);
		        RegSetValueEx(hkey, PRINTERRORKEY, 0, REG_DWORD,
			    (CONST BYTE *)&printerror, sizeof(DWORD));
		    }
		    RegCloseKey(hkey);
		}
		}
		EndDialog(hDlg, TRUE);
		return(TRUE);
	      case IDCANCEL:
                EndDialog(hDlg, FALSE);
                return(TRUE);
              default:
                return(FALSE);
            }
        default:
            return(FALSE);
    }
}

/* return length of env in characters (which may not be bytes) */
int
env_length(LPTSTR env)
{
LPTSTR p;
    p = env;
    while (*p) {
	while (*p)
	    p++;
	p++;
    }
    p++;
    return (p - env);
}

/* join two environment blocks together, returning a handle
 * to a new environment block.  The caller must free this with
 * GlobalFree() when it is finished.
 */
HGLOBAL
join_env(LPTSTR env1, LPTSTR env2)
{
int len1, len2;
HGLOBAL henv;
LPTSTR env;
    len1 = env_length(env1);	/* length in characters, not bytes */
    len2 = env_length(env2);
    henv = GlobalAlloc(GPTR, (DWORD)((len1 + len2)*sizeof(TCHAR)));  
    env = GlobalLock(henv);
    if (env == NULL)
	return NULL;
    MoveMemory(env, env1, len1*sizeof(TCHAR));
    env += len1 - 1;
    MoveMemory(env, env2, len2*sizeof(TCHAR)); 
    GlobalUnlock(henv);
    return henv;
}

/* Append a string to an environment block.
 * It is assumed that the environment block has sufficient
 * space.
 *  env is the environment block.
 *  name is the environment variable name, which includes a trailing '='.
 *  len is the length of name in bytes, including the trailing null.
 *  value is the environment variable value.
 */
void
append_env(LPTSTR env, LPTSTR name, int len, LPTSTR value)
{
int oldlen;
  oldlen = env_length(env);
  env = env + oldlen - 1;
  MoveMemory(env, name, len);
  env += len/sizeof(TCHAR) - 1;
  MoveMemory(env, value, lstrlen(value)*sizeof(TCHAR));
  env += lstrlen(value) + 1;
  *env = '\0';
}

/* create an environment variable block which contains
 * some RedMon extras about the print job.
 */
HGLOBAL make_job_env(JOB_INFO_1 * ji1, LPTSTR port)
{
int len;
TCHAR buf[32];
TCHAR temp[256];
HGLOBAL henv;
LPTSTR env;
    wsprintf(buf, TEXT("%d"), ji1->JobId);
    get_temp(temp, sizeof(temp)/sizeof(TCHAR));
    len = sizeof(REDMON_PORT) + 
          sizeof(REDMON_JOB) + 
          sizeof(REDMON_PRINTER) + 
          sizeof(REDMON_MACHINE) + 
          sizeof(REDMON_USER) + 
          sizeof(REDMON_DOCNAME) + 
          sizeof(REDMON_TEMP) + 
          sizeof(REDMON_TMP) + 
          (lstrlen(port) + 
          lstrlen(buf) + 
          lstrlen(ji1->pPrinterName) + 
          lstrlen(ji1->pMachineName) + 
          lstrlen(ji1->pUserName) + 
          lstrlen(ji1->pDocument) +
          lstrlen(temp) +
          lstrlen(temp) + 1) * sizeof(TCHAR);
    henv = GlobalAlloc(GPTR, len);
    env = GlobalLock(henv);
    if (env == NULL)
	return NULL;
    append_env(env, REDMON_PORT, sizeof(REDMON_PORT), port);
    append_env(env, REDMON_JOB, sizeof(REDMON_JOB), buf);
    append_env(env, REDMON_PRINTER, sizeof(REDMON_PRINTER), ji1->pPrinterName);
    append_env(env, REDMON_MACHINE, sizeof(REDMON_MACHINE), ji1->pMachineName);
    append_env(env, REDMON_USER, sizeof(REDMON_USER), ji1->pUserName);
    append_env(env, REDMON_DOCNAME, sizeof(REDMON_DOCNAME), ji1->pDocument);
    append_env(env, REDMON_TEMP, sizeof(REDMON_TEMP), temp);
    append_env(env, REDMON_TMP, sizeof(REDMON_TMP), temp);
    GlobalUnlock(henv);
    return henv;
}



#ifdef UNUSED
/* print to stdout the contents of the environment block */
void
dump_env(REDATA *prd, LPTSTR env)
{
TCHAR  buf[1024];
LPTSTR name, next;
    next = env;
    while (*next) {
        name = next;
	while (*next)
	    next++;
	wsprintf(buf, TEXT("%s\n"), name);
	write_string_to_log(prd, buf);
	next++;
    }
}
#endif


BOOL make_env(REDATA * prd)
{
LPTSTR env, extra_env;
HGLOBAL h_extra_env;
HGLOBAL hPrinter;
    if (OpenPrinter(prd->pPrinterName, &hPrinter, NULL)) {
	DWORD dwNeeded = 0;
	HGLOBAL hglobal = GlobalAlloc(GPTR, (DWORD)4096);
	JOB_INFO_1 *pjob = (JOB_INFO_1 *)GlobalLock(hglobal);
	if ((pjob != (JOB_INFO_1 *)NULL) && 
	     GetJob(hPrinter, prd->JobId, 1, (LPBYTE)pjob, 
		4096, &dwNeeded)) {

	    /* Add some environment variables */
	    /* It would be simpler to use SetEnvironmentVariable()
	     * and then GetEnvironmentStrings(), then to delete
	     * the environment variables we added, but this could
	     * cause problems if two RedMon ports simultaneously
	     * did this.
	     */
	    env = GetEnvironmentStrings();
	    h_extra_env= make_job_env(pjob, prd->portname);
	    extra_env = GlobalLock(h_extra_env);
	    prd->environment = join_env(env, extra_env);
	    GlobalUnlock(prd->environment);
	    GlobalUnlock(h_extra_env);
	    GlobalFree(h_extra_env);
	    FreeEnvironmentStrings(env);

	}
	if (pjob != (JOB_INFO_1 *)NULL) {
	    GlobalUnlock(hglobal);
	    GlobalFree(hglobal);
	}
	ClosePrinter(hPrinter);
	return TRUE;
    }
    return FALSE;
}

/* start_redirect() was originally based on an example in the Win32 SDK
 * which used GetStdHandle() and SetStdHandle() to redirect stdio.
 * The example works under Windows 95, but not under NT.
 * For NT, we need to use 
 *  siStartInfo.dwFlags = STARTF_USESTDHANDLES;
 *  siStartInfo.hStdInput = prd->hChildStdinRd;
 *  siStartInfo.hStdOutput = prd->hChildStdoutWr;
 *  siStartInfo.hStdError = prd->hChildStderrWr;
 * The SDK example does NOT include these.  Most strange for an 
 * example that was written before Windows 95 existed!
 * STARTF_USESTDHANDLES also works for Windows 95, so the original 
 * code is commented out with #ifdef SAVESTD / #endif
 */

/* Start child program with redirected standard input and output */
BOOL start_redirect(REDATA * prd)
{
    SECURITY_ATTRIBUTES saAttr;
    STARTUPINFO siStartInfo;
    HANDLE hPipeTemp;
    LPVOID env;

    /* Set the bInheritHandle flag so pipe handles are inherited. */
    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = NULL;

#ifdef SAVESTD
    /* Save the current standard handles . */
    prd->hSaveStdout = GetStdHandle(STD_OUTPUT_HANDLE);
    prd->hSaveStderr = GetStdHandle(STD_ERROR_HANDLE);
    prd->hSaveStdin = GetStdHandle(STD_INPUT_HANDLE);
#endif

    /* Create anonymous inheritable pipes for STDIN, STDOUT and STDERR
     * for child. For each pipe, create a noninheritable duplicate handle
     * of our end of the pipe, then close the inheritable handle.
     */
    if (!CreatePipe(&prd->hChildStdinRd, &hPipeTemp, &saAttr, 0))
	return FALSE;
    if (!DuplicateHandle(GetCurrentProcess(), hPipeTemp,
            GetCurrentProcess(), &prd->hChildStdinWr, 0,
            FALSE,       /* not inherited */
            DUPLICATE_SAME_ACCESS)) {
        CloseHandle(hPipeTemp);
	return FALSE;
    }
    CloseHandle(hPipeTemp);

    if (!CreatePipe(&hPipeTemp, &prd->hChildStdoutWr, &saAttr, 0))
	return FALSE;	/* cleanup of pipes will occur in caller */
    if (!DuplicateHandle(GetCurrentProcess(), hPipeTemp,
            GetCurrentProcess(), &prd->hChildStdoutRd, 0,
            FALSE,       /* not inherited */
            DUPLICATE_SAME_ACCESS)) {
        CloseHandle(hPipeTemp);
	return FALSE;
    }
    CloseHandle(hPipeTemp);

    if (!CreatePipe(&hPipeTemp, &prd->hChildStderrWr, &saAttr, 0))
	return FALSE;
    if (!DuplicateHandle(GetCurrentProcess(), hPipeTemp,
            GetCurrentProcess(), &prd->hChildStderrRd, 0,
            FALSE,       /* not inherited */
            DUPLICATE_SAME_ACCESS)) {
        CloseHandle(hPipeTemp);
	return FALSE;
    }
    CloseHandle(hPipeTemp);

#ifdef SAVESTD
    if (!SetStdHandle(STD_OUTPUT_HANDLE, prd->hChildStdoutWr))
	return FALSE;
    if (!SetStdHandle(STD_ERROR_HANDLE, prd->hChildStderrWr))
	return FALSE;
    if (!SetStdHandle(STD_INPUT_HANDLE, prd->hChildStdinRd))
	return FALSE;
#endif

    /* Now create the child process. */

    /* Set up members of STARTUPINFO structure. */

    siStartInfo.cb = sizeof(STARTUPINFO);
    siStartInfo.lpReserved = NULL;
    siStartInfo.lpDesktop = NULL;
    siStartInfo.lpTitle = NULL;  /* use executable name as title */
    siStartInfo.dwX = siStartInfo.dwY = CW_USEDEFAULT;		/* ignored */
    siStartInfo.dwXSize = siStartInfo.dwYSize = CW_USEDEFAULT;	/* ignored */
    siStartInfo.dwXCountChars = 80;
    siStartInfo.dwYCountChars = 25;
    siStartInfo.dwFillAttribute = 0;			/* ignored */
    siStartInfo.dwFlags = STARTF_USESTDHANDLES;
    siStartInfo.wShowWindow = SW_SHOWNORMAL;		/* ignored */
    if (prd->show != SHOW_NORMAL)
        siStartInfo.dwFlags |= STARTF_USESHOWWINDOW;
    if (prd->show == SHOW_MIN)
        siStartInfo.wShowWindow = SW_SHOWMINNOACTIVE;	
    else if (prd->show == SHOW_HIDE)
        siStartInfo.wShowWindow = SW_HIDE;	
    siStartInfo.cbReserved2 = 0;
    siStartInfo.lpReserved2 = NULL;
    siStartInfo.hStdInput = prd->hChildStdinRd;
    siStartInfo.hStdOutput = prd->hChildStdoutWr;
    siStartInfo.hStdError = prd->hChildStderrWr;

    if (prd->environment)
        env = GlobalLock(prd->environment);
    else
	env = NULL;

    /* Create the child process. */

#if defined(UNICODE) && !defined(NT35) && !defined(__BORLANDC__)
    if (prd->runuser) {
	BOOL flag;
	HANDLE htoken;
	TCHAR buf[MAXSTR];
        /* CreateProcessAsUser is only supported by NT3.51 and later. */
	/* get impersonation token of current thread */
	if ( !(flag = OpenThreadToken(GetCurrentThread() , 
	    TOKEN_DUPLICATE | TOKEN_IMPERSONATE, 
	    TRUE, &htoken)) ) {
	    DWORD err = GetLastError();
	    wsprintf(buf, TEXT("OpenThreadToken failed, error code=%d\r\n"),
		err);
	    write_string_to_log(prd, buf);
	    write_error(prd, err);
	}
	/* Duplicate it to create a primary token */
	if ( !(flag = DuplicateTokenEx(htoken, TOKEN_ALL_ACCESS, NULL, 
	       SecurityImpersonation, TokenPrimary, &prd->primary_token)) ) {
	    DWORD err = GetLastError();
	    wsprintf(buf, TEXT("DuplicateTokenEx failed, error code=%d\r\n"),
		err);
	    write_string_to_log(prd, buf);
	    write_error(prd, err);
	}

	/* Be default, the process is created on an invisible desktop.
	 * that can't receive input.
	 * Instead we try to put the process on the main desktop.
	 * We need to change the DACL of the desktop, but I don't know
	 * how to do this yet, so the desktop code is disabled.
	 */
	/*
        siStartInfo.lpDesktop = "winsta0\\desktop";
	*/
	if ( !(flag = CreateProcessAsUser(prd->primary_token, NULL,
		prd->command,  /* command line                       */
		NULL,          /* process security attributes        */
		NULL,          /* primary thread security attributes */
		TRUE,          /* handles are inherited              */
		CREATE_UNICODE_ENVIRONMENT,  /* creation flags       */
		env,           /* environment                        */
		NULL,          /* use parent's current directory     */
		&siStartInfo,  /* STARTUPINFO pointer                */
		&prd->piProcInfo))  /* receives PROCESS_INFORMATION  */
	   ) {
	    DWORD err = GetLastError();
	    wsprintf(buf, TEXT("CreateProcessAsUser failed, error code=%d\r\n"),
		err);
	    write_string_to_log(prd, buf);
	    write_error(prd, err);
	}
	
	if (!flag)
	   prd->runuser = FALSE;
    }
    if (!prd->runuser)
#endif

    if (!CreateProcess(NULL,
        prd->command,  /* command line                       */
        NULL,          /* process security attributes        */
        NULL,          /* primary thread security attributes */
        TRUE,          /* handles are inherited              */
#ifdef UNICODE
	CREATE_UNICODE_ENVIRONMENT,  /* creation flags       */
#else
	0,             /* creation flags                     */
#endif
	env,           /* environment                        */
        NULL,          /* use parent's current directory     */
        &siStartInfo,  /* STARTUPINFO pointer                */
        &prd->piProcInfo))  /* receives PROCESS_INFORMATION  */
	  return FALSE;

    /* After process creation, restore the saved STDIN and STDOUT. */

#ifdef SAVESTD
    if (!SetStdHandle(STD_INPUT_HANDLE, prd->hSaveStdin))
	return FALSE;
    if (!SetStdHandle(STD_OUTPUT_HANDLE, prd->hSaveStdout))
	return FALSE;
    if (!SetStdHandle(STD_ERROR_HANDLE, prd->hSaveStderr))
	return FALSE;
#endif

    return TRUE;
}



#ifdef MAKEEXE
BYTE pPorts[4096];
DWORD cbBuf, cbNeeded, cReturned;
TCHAR mess[MAXSTR];
#define QUICKBROWN "The quick brown fox jumps over the lazy dog."
/*
#define PSTEST "(c:/rjl/monitor/test.txt) (w) file\n dup (Hello, world\\n) writestring closefile\n(Hello, world\\n) print flush\n"
#define PSTEST "(c:/gstools/gs5.50/colorcir.ps) run flush\n"
*/
#define PSTEST "(d:/data/src/a20.ps) run flush\n"

int PASCAL 
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int cmdShow)
{
    DWORD written;
    HANDLE hport;
    PORT_INFO_2 *pi2;
    PORT_INFO_1 *pi1;
    DOC_INFO_1 dci1;
    int i;
#ifdef UNICODE
    LPMONITOREX pmon;
#else
    MONITOR mon;
#endif
    TCHAR monitorname[256];
    TCHAR mess[256];
    hdll = hInstance;
#ifdef UNICODE
    pmon = InitializePrintMonitor(TEXT("System\\CurrentControlSet\\control\\Print\\Monitors\\Redirected Port"));
#else
    InitializeMonitorEx(TEXT("System\\CurrentControlSet\\control\\Print\\Monitors\\Redirected Port"), &mon);
#endif

    LoadString(hdll, IDS_MONITORNAME, monitorname, 
	sizeof(monitorname)/sizeof(TCHAR)-1);

#ifdef UNUSED
    rAddPort(NULL, HWND_DESKTOP, monitorname);
    rAddPort(NULL, HWND_DESKTOP, monitorname);
#endif
    rAddPort(NULL, HWND_DESKTOP, monitorname);

#ifdef UNICODE
    pi2 = (PORT_INFO_2 *)malloc(sizeof(PORT_INFO_2));
    pi2->pPortName = TEXT("RPT7:");
    pi2->pMonitorName = MONITORNAME;
    pi2->pDescription = TEXT("Redirected Port 7");
    pi2->fPortType = PORT_TYPE_WRITE;
    pi2->Reserved = 0;
    rAddPortEx(NULL, 2, pi2, TEXT("Redirected Port"));
/*
    rDeletePort(NULL, HWND_DESKTOP, TEXT("RPT7:"));
*/

    pi1 = (PORT_INFO_1 *)malloc(sizeof(PORT_INFO_1));
    pi1->pName = TEXT("RPT8:");
    rAddPortEx(NULL, 1, pi1, TEXT("Redirected Port"));
    rDeletePort(NULL, HWND_DESKTOP, TEXT("RPT8:"));
    free(pi2);
#endif
    rConfigurePort(NULL, HWND_DESKTOP, TEXT("RPT1:"));

    cbBuf = 0;
    rEnumPorts(NULL, 2, pPorts, cbBuf, &cbNeeded, &cReturned);
    cbBuf = cbNeeded;
    rEnumPorts(NULL, 2, pPorts, cbBuf, &cbNeeded, &cReturned);
    pi2 = (PORT_INFO_2 *)pPorts;

    for (i=0; i<cReturned; i++) {
	wsprintf(mess, TEXT("\042%s\042 \042%s\042 \042%s\042 %d\n"), 
	   pi2[i].pPortName,
	   pi2[i].pMonitorName,
	   pi2[i].pDescription,
	   pi2[i].fPortType
	   );
        MessageBox(NULL, mess, TEXT("EXE for Redirect Monitor"), MB_OK);
    }
#ifdef UNUSED
#endif

    rOpenPort(TEXT("RPT1:"), &hport);
    dci1.pDocName = TEXT("Test document");
    dci1.pOutputFile = NULL;
    dci1.pDatatype = NULL;
#ifdef UNICODE
    rStartDocPort(hport, TEXT("Apple LaserWriter II NT v47.0"), 1, 1, (LPBYTE)&dci1);
#else
    rStartDocPort(hport, TEXT("Apple LaserWriter II NT"), 1, 1, (LPBYTE)&dci1);
#endif
//    rWritePort(hport, (LPBYTE)&QUICKBROWN, strlen(QUICKBROWN), &written);
    rWritePort(hport, (LPBYTE)&PSTEST, strlen(PSTEST), &written);
    rWritePort(hport, (LPBYTE)&PSTEST, strlen(PSTEST), &written);
    rEndDocPort(hport);
    rClosePort(hport);

    rDeletePort(NULL, HWND_DESKTOP, TEXT("RPT2:"));

    return 0;
}
#endif

/* end of redmon.c */
